from petsc4py import PETSc
import unittest


def mkmat(n, mtype, opts):
    A = PETSc.Mat().create(PETSc.COMM_SELF)
    A.setSizes([n, n])
    A.setType(mtype)
    A.setUp()
    for o in opts:
        A.setOption(o, True)
    return A


def mksys_diag(n, mtype, opts):
    A = mkmat(n, mtype, opts)
    x, b = A.createVecs()
    for i in range(n):
        A[i, i] = i + 1
        x[i] = 1.0 / (i + 1)
        b[i] = 1
    A.assemble()
    x.assemble()
    b.assemble()
    return A, x, b


def mksys_poi2(n, mtype, opts):
    A = mkmat(n, mtype, opts)
    x, b = A.createVecs()
    for i in range(n):
        if i == 0:
            cols = [i, i + 1]
            vals = [2, -1]
        elif i == n - 1:
            cols = [i - 1, i]
            vals = [-1, 2]
        else:
            cols = [i - 1, i, i + 1]
            vals = [-1, 2, -1]
        A[i, cols] = vals
        x[i] = i + 1
        b[i] = 0
    A.assemble()
    x.assemble()
    b.assemble()
    A.mult(x, b)
    return A, x, b


class BaseTestMatFactor:
    MKSYS = None
    MTYPE = None
    MOPTS = ()

    def setUp(self):
        A, x, b = self.MKSYS(10, self.MTYPE, self.MOPTS)
        self.A = A
        self.x = x
        self.b = b

    def tearDown(self):
        self.A.setUnfactored()
        self.A.destroy()
        self.A = None
        self.x.destroy()
        self.x = None
        self.b.destroy()
        self.b = None
        PETSc.garbage_cleanup()


class BaseTestMatFactorLU(BaseTestMatFactor):
    def testFactorLU(self):
        r, c = self.A.getOrdering('nd')
        self.A.reorderForNonzeroDiagonal(r, c)
        self.A.factorLU(r, c, {'zeropivot': 1e-5})
        x = self.x.duplicate()
        self.A.solve(self.b, x)
        x.axpy(-1, self.x)
        self.assertTrue(x.norm() < 1e-3)


class BaseTestMatFactorILU(BaseTestMatFactor):
    def testFactorILU(self):
        r, c = self.A.getOrdering('natural')
        self.A.factorILU(r, c, {'levels': 0})
        x = self.x.duplicate()
        self.A.solve(self.b, x)
        x.axpy(-1, self.x)
        self.assertTrue(x.norm() < 1e-3)


## class BaseTestMatFactorILUDT(BaseTestMatFactor):
##
##     def testFactorILUDT(self):
##         r, c = self.A.getOrdering("natural")
##         self.A = self.A.factorILUDT(r,c)
##         x = self.x.duplicate()
##         self.A.solve(self.b, x)
##         x.axpy(-1, self.x)
##         self.assertTrue(x.norm() < 1e-3)
##
class BaseTestMatFactorChol(BaseTestMatFactor):
    def testFactorChol(self):
        r, c = self.A.getOrdering('natural')
        self.A.factorCholesky(r)
        x = self.x.duplicate()
        self.A.solve(self.b, x)
        x.axpy(-1, self.x)
        self.assertTrue(x.norm() < 1e-3)


class BaseTestMatFactorICC(BaseTestMatFactor):
    def testFactorICC(self):
        r, c = self.A.getOrdering('natural')
        self.A.factorICC(r)
        x = self.x.duplicate()
        self.A.solve(self.b, x)
        x.axpy(-1, self.x)
        self.assertTrue(x.norm() < 1e-3)


# --------------------------------------------------------------------


class TestMatFactorA1(BaseTestMatFactorLU, BaseTestMatFactorChol, unittest.TestCase):
    MKSYS = staticmethod(mksys_diag)
    MTYPE = PETSc.Mat.Type.SEQDENSE


class TestMatFactorA2(BaseTestMatFactorLU, BaseTestMatFactorChol, unittest.TestCase):
    MKSYS = staticmethod(mksys_poi2)
    MTYPE = PETSc.Mat.Type.SEQDENSE


# ---


class TestMatFactorB1(
    BaseTestMatFactorLU,
    BaseTestMatFactorILU,
    ## BaseTestMatFactorILUDT,
    unittest.TestCase,
):
    MKSYS = staticmethod(mksys_diag)
    MTYPE = PETSc.Mat.Type.SEQAIJ


class TestMatFactorB2(
    BaseTestMatFactorLU,
    BaseTestMatFactorILU,
    ## BaseTestMatFactorILUDT,
    unittest.TestCase,
):
    MKSYS = staticmethod(mksys_poi2)
    MTYPE = PETSc.Mat.Type.SEQAIJ


# ---


class TestMatFactorC1(BaseTestMatFactorLU, BaseTestMatFactorILU, unittest.TestCase):
    MKSYS = staticmethod(mksys_diag)
    MTYPE = PETSc.Mat.Type.SEQBAIJ


class TestMatFactorC2(BaseTestMatFactorLU, BaseTestMatFactorILU, unittest.TestCase):
    MKSYS = staticmethod(mksys_poi2)
    MTYPE = PETSc.Mat.Type.SEQBAIJ


# ---


class TestMatFactorD1(BaseTestMatFactorChol, BaseTestMatFactorICC, unittest.TestCase):
    MKSYS = staticmethod(mksys_diag)
    MTYPE = PETSc.Mat.Type.SEQSBAIJ
    MOPTS = [PETSc.Mat.Option.IGNORE_LOWER_TRIANGULAR]


class TestMatFactorD2(BaseTestMatFactorChol, BaseTestMatFactorICC, unittest.TestCase):
    MKSYS = staticmethod(mksys_poi2)
    MTYPE = PETSc.Mat.Type.SEQSBAIJ
    MOPTS = [PETSc.Mat.Option.IGNORE_LOWER_TRIANGULAR]


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()



from petsc4py.PETSc cimport Vec,  PetscVec
from petsc4py.PETSc cimport Mat,  PetscMat
from petsc4py.PETSc cimport DM,   PetscDM
from petsc4py.PETSc cimport SNES

from petsc4py.PETSc import Error

cdef extern from "Bratu3Dimpl.h":
    ctypedef struct Params:
        double lambda_
    int FormInitGuess(PetscDM da, PetscVec x, Params *p)
    int FormFunction (PetscDM da, PetscVec x, PetscVec F, Params *p)
    int FormJacobian (PetscDM da, PetscVec x, PetscMat J, Params *p)


def formInitGuess(Vec x, DM da, double lambda_):
    cdef int ierr
    cdef Params p = {"lambda_" : lambda_}
    ierr = FormInitGuess(da.dm, x.vec, &p)
    if ierr != 0: raise Error(ierr)


def formFunction(SNES snes, Vec x, Vec f, DM da, double lambda_):
    cdef int ierr
    cdef Params p = {"lambda_" : lambda_}
    ierr = FormFunction(da.dm, x.vec, f.vec, &p)
    if ierr != 0: raise Error(ierr)


def formJacobian(SNES snes, Vec x, Mat J, Mat P, DM da, double lambda_):
    cdef int ierr
    cdef Params p = {"lambda_" : lambda_}
    ierr = FormJacobian(da.dm, x.vec, P.mat, &p)
    if ierr != 0: raise Error(ierr)
    if J != P: J.assemble() # for matrix-free operator
    return Mat.Structure.SAME_NONZERO_PATTERN



#!/usr/bin/env python3

# --------------------------------------------------------------------


# --------------------------------------------------------------------

try:
    from docutils.nodes import NodeVisitor

    NodeVisitor.unknown_visit = lambda self, node: None
    NodeVisitor.unknown_departure = lambda self, node: None
except ImportError:
    pass

try:  # epydoc 3.0.1 + docutils 0.6
    from docutils.nodes import Text
    from UserString import UserString

    if not isinstance(Text, UserString):

        def Text_get_data(s):
            try:
                return s._data
            except AttributeError:
                return s.astext()

        def Text_set_data(s, d):
            s.astext = lambda: d
            s._data = d

        Text.data = property(Text_get_data, Text_set_data)
except ImportError:
    pass

# --------------------------------------------------------------------

from epydoc.docwriter import dotgraph
from epydoc import docstringparser as dsp

import re
import sys
import os
import epydoc.cli

dotgraph._DOT_VERSION_RE = re.compile(r'dot (?:- Graphviz )version ([\d\.]+)')

try:
    dotgraph.DotGraph.DEFAULT_HTML_IMAGE_FORMAT = 'png'

except AttributeError:
    DotGraph_to_html = dotgraph.DotGraph.to_html
    DotGraph_run_dot = dotgraph.DotGraph._run_dot

    def to_html(self, image_file, image_url, center=True):
        if image_file[-4:] == '.gif':
            image_file = image_file[:-4] + '.png'
        if image_url[-4:] == '.gif':
            image_url = image_url[:-4] + '.png'
        return DotGraph_to_html(self, image_file, image_url)

    def _run_dot(self, *options):
        if '-Tgif' in options:
            opts = list(options)
            for i, o in enumerate(opts):
                if o == '-Tgif':
                    opts[i] = '-Tpng'
            options = type(options)(opts)
        return DotGraph_run_dot(self, *options)

    dotgraph.DotGraph.to_html = to_html
    dotgraph.DotGraph._run_dot = _run_dot

# --------------------------------------------------------------------


_SIGNATURE_RE = re.compile(
    # Class name (for builtin methods)
    r'^\s*((?P<class>\w+)\.)?'
    +
    # The function name
    r'(?P<func>\w+)'
    +
    # The parameters
    r'\(((?P<self>(?:self|cls|mcs)),?)?(?P<params>.*)\)'
    +
    # The return value (optional)
    r'(\s*(->)\s*(?P<return>\S.*?))?'
    +
    # The end marker
    r'\s*(\n|\s+(--|<=+>)\s+|$|\.\s+|\.\n)'
)


dsp._SIGNATURE_RE = _SIGNATURE_RE

# --------------------------------------------------------------------


def epydocify():
    dirname = os.path.dirname(__file__)
    config = os.path.join(dirname, 'epydoc.cfg')
    sys.argv.append('--config=' + config)
    epydoc.cli.cli()


if __name__ == '__main__':
    epydocify()

# --------------------------------------------------------------------



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""The PETSc for Python package.

This package is an interface to PETSc libraries.

PETSc_ (the Portable, Extensible Toolkit for Scientific Computation)
is a suite of data structures and routines for the scalable (parallel)
solution of scientific applications modeled by partial differential
equations. It employs the MPI_ standard for all message-passing
communications.

.. _PETSc: https://petsc.org
.. _MPI:   https://www.mpi-forum.org

"""

__author__ = 'Lisandro Dalcin'
__version__ = '3.24.6'
__credits__ = 'PETSc Team <petsc-maint@mcs.anl.gov>'


def init(args=None, arch=None, comm=None):
    """Initialize PETSc.

    Parameters
    ----------
    args
        Command-line arguments, usually the `sys.argv` list
    arch
        Specific configuration to use
    comm
        MPI commmunicator

    Notes
    -----
    This function should be called only once, typically at the very
    beginning of the bootstrap script of an application.
    """
    import petsc4py.lib

    PETSc = petsc4py.lib.ImportPETSc(arch)
    args = petsc4py.lib.getInitArgs(args)
    PETSc._initialize(args, comm)


def get_include():
    """Return the directory in the package that contains header files.

    Extension modules that need to compile against petsc4py should use
    this function to locate the appropriate include directory.

    Example
    -------
    Using Python distutils or NumPy distutils::

        import petsc4py
        Extension('extension_name', ...
                  include_dirs=[..., petsc4py.get_include()])

    """
    from os.path import dirname, join

    return join(dirname(__file__), 'include')


def get_config():
    """Return a dictionary with information about PETSc."""
    import os

    from io import StringIO
    from configparser import ConfigParser
    pgkdir = os.path.dirname(__file__)
    filename = os.path.join(pgkdir, 'lib', 'petsc.cfg')
    with open(filename) as fp:
        stream = StringIO('[petsc]\n' + fp.read())
    parser = ConfigParser()
    parser.optionxform = str
    parser.read_file(stream, filename)
    return dict(parser.items('petsc'))



# cython: language_level=3str
# cython: embedsignature=True
# cython: embedsignature.format=python
# cython: annotation_typing=False
# cython: cdivision=True
# cython: auto_pickle=False
# cython: always_allow_keywords=True
# cython: allow_none_for_extension_args=False
# cython: autotestdict=False
# cython: warn.multiple_declarators=False
# cython: optimize.use_switch=False
# cython: binding=True

from __future__ import annotations
cimport cython # no-cython-lint

include "PETSc/PETSc.pyx"



# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef enum int "PetscGaussLobattoLegendreCreateType":
        PETSCGAUSSLOBATTOLEGENDRE_VIA_LINEAR_ALGEBRA
        PETSCGAUSSLOBATTOLEGENDRE_VIA_NEWTON

    PetscErrorCode PetscFECreateDefault(MPI_Comm, PetscInt, PetscInt, PetscBool, const char [], PetscInt, PetscFE*)
    PetscErrorCode PetscQuadratureCreate(MPI_Comm, PetscQuadrature*)
    PetscErrorCode PetscQuadratureDuplicate(PetscQuadrature, PetscQuadrature*)
    PetscErrorCode PetscQuadratureGetOrder(PetscQuadrature, PetscInt*)
    PetscErrorCode PetscQuadratureSetOrder(PetscQuadrature, PetscInt)
    PetscErrorCode PetscQuadratureGetNumComponents(PetscQuadrature, PetscInt*)
    PetscErrorCode PetscQuadratureSetNumComponents(PetscQuadrature, PetscInt)
    PetscErrorCode PetscQuadratureGetData(PetscQuadrature, PetscInt*, PetscInt*, PetscInt*, const PetscReal *[], const PetscReal *[])
    PetscErrorCode PetscQuadratureSetData(PetscQuadrature, PetscInt, PetscInt, PetscInt, const PetscReal [], const PetscReal [])

    PetscErrorCode PetscQuadratureView(PetscQuadrature, PetscViewer)
    PetscErrorCode PetscQuadratureDestroy(PetscQuadrature *)

    PetscErrorCode PetscQuadratureExpandComposite(PetscQuadrature, PetscInt, const PetscReal[], const PetscReal[], PetscQuadrature *)

    PetscErrorCode PetscDTLegendreEval(PetscInt, const PetscReal*, PetscInt, const PetscInt*, PetscReal*, PetscReal*, PetscReal*)
    PetscErrorCode PetscDTGaussQuadrature(PetscInt, PetscReal, PetscReal, PetscReal*, PetscReal*)
    PetscErrorCode PetscDTGaussLobattoLegendreQuadrature(PetscInt, PetscGaussLobattoLegendreCreateType, PetscReal*, PetscReal*)
    PetscErrorCode PetscDTReconstructPoly(PetscInt, PetscInt, const PetscReal*, PetscInt, const PetscReal*, PetscReal*)
    PetscErrorCode PetscDTGaussTensorQuadrature(PetscInt, PetscInt, PetscInt, PetscReal, PetscReal, PetscQuadrature*)
    PetscErrorCode PetscDTGaussJacobiQuadrature(PetscInt, PetscInt, PetscInt, PetscReal, PetscReal, PetscQuadrature*)

    PetscErrorCode PetscDTTanhSinhTensorQuadrature(PetscInt, PetscInt, PetscReal, PetscReal, PetscQuadrature *)
    PetscErrorCode PetscDTTanhSinhIntegrate(void (*)(PetscReal *, void *, PetscReal *), PetscReal, PetscReal, PetscInt, PetscReal *)
    PetscErrorCode PetscDTTanhSinhIntegrateMPFR(void (*)(PetscReal *, void *, PetscReal *), PetscReal, PetscReal, PetscInt, PetscReal *)

    PetscErrorCode PetscGaussLobattoLegendreIntegrate(PetscInt, PetscReal *, PetscReal *, const PetscReal *, PetscReal *)
    PetscErrorCode PetscGaussLobattoLegendreElementLaplacianCreate(PetscInt, PetscReal *, PetscReal *, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementLaplacianDestroy(PetscInt, PetscReal *, PetscReal *, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementGradientCreate(PetscInt, PetscReal *, PetscReal *, PetscReal ***, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementGradientDestroy(PetscInt, PetscReal *, PetscReal *, PetscReal ***, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementAdvectionCreate(PetscInt, PetscReal *, PetscReal *, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementAdvectionDestroy(PetscInt, PetscReal *, PetscReal *, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementMassCreate(PetscInt, PetscReal *, PetscReal *, PetscReal ***)
    PetscErrorCode PetscGaussLobattoLegendreElementMassDestroy(PetscInt, PetscReal *, PetscReal *, PetscReal ***)



# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef int PetscClassId
    ctypedef long PetscObjectState
    ctypedef long PetscObjectId
    PetscErrorCode PetscObjectView(PetscObject, PetscViewer)
    PetscErrorCode PetscObjectDestroy(PetscObject*)
    PetscErrorCode PetscObjectGetReference(PetscObject, PetscInt*)
    PetscErrorCode PetscObjectReference(PetscObject)
    PetscErrorCode PetscObjectDereference(PetscObject)

    PetscErrorCode PetscObjectSetOptionsPrefix(PetscObject, char[])
    PetscErrorCode PetscObjectAppendOptionsPrefix(PetscObject, char[])
    PetscErrorCode PetscObjectGetOptionsPrefix(PetscObject, char*[])
    PetscErrorCode PetscObjectSetFromOptions(PetscObject)
    PetscErrorCode PetscObjectViewFromOptions(PetscObject, PetscObject, char[])

    PetscErrorCode PetscObjectGetComm(PetscObject, MPI_Comm*)
    PetscErrorCode PetscObjectGetClassId(PetscObject, PetscClassId*)
    PetscErrorCode PetscObjectGetType(PetscObject, char*[])
    PetscErrorCode PetscObjectGetClassName(PetscObject, char*[])
    PetscErrorCode PetscObjectSetName(PetscObject, char[])
    PetscErrorCode PetscObjectGetName(PetscObject, char*[])
    PetscErrorCode PetscObjectGetId(PetscObject, PetscObjectId*)

    PetscErrorCode PetscObjectStateIncrease(PetscObject)
    PetscErrorCode PetscObjectStateSet(PetscObject, PetscObjectState)
    PetscErrorCode PetscObjectStateGet(PetscObject, PetscObjectState*)
    PetscErrorCode PetscObjectTypeCompare(PetscObject, char[], PetscBool*)
    PetscErrorCode PetscObjectChangeTypeName(PetscObject, char[])
    PetscErrorCode PetscObjectCompose(PetscObject, char[], PetscObject)
    PetscErrorCode PetscObjectQuery(PetscObject, char[], PetscObject*)

    ctypedef void (*PetscVoidFunction)()
    PetscErrorCode PetscObjectComposeFunction(PetscObject, char[], PetscVoidFunction)
    PetscErrorCode PetscObjectQueryFunction(PetscObject, char[], PetscVoidFunction*)

    PetscErrorCode PetscObjectIncrementTabLevel(PetscObject, PetscObject, PetscInt)
    PetscErrorCode PetscObjectGetTabLevel(PetscObject, PetscInt*)
    PetscErrorCode PetscObjectSetTabLevel(PetscObject, PetscInt)

    ctypedef struct _n_PetscOptionItems
    ctypedef _n_PetscOptionItems* PetscOptionItems

    ctypedef PetscErrorCode (*PetscObjectOptionsHandler)(PetscObject,
                                                         PetscOptionItems,
                                                         void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscObjectOptionsCtxDel)(PetscObject, void*)
    PetscErrorCode PetscObjectAddOptionsHandler(PetscObject,
                                                PetscObjectOptionsHandler,
                                                PetscObjectOptionsCtxDel,
                                                void*)
    PetscErrorCode PetscObjectDestroyOptionsHandlers(PetscObject)

cdef extern from * nogil: # custom.h
    PetscErrorCode PetscObjectGetDeviceId(PetscObject, PetscInt*)

cdef extern from "<petsc/private/garbagecollector.h>" nogil:
    PetscErrorCode PetscObjectDelayedDestroy(PetscObject*)

# --------------------------------------------------------------------

cdef inline PetscErrorCode PetscINCREF(PetscObject *obj) noexcept nogil:
    if obj    == NULL: return PETSC_SUCCESS
    if obj[0] == NULL: return PETSC_SUCCESS
    return PetscObjectReference(obj[0])

cdef inline PetscErrorCode PetscDECREF(PetscObject *obj) noexcept nogil:
    if obj    == NULL: return PETSC_SUCCESS
    if obj[0] == NULL: return PETSC_SUCCESS
    return PetscObjectDereference(obj[0])

cdef inline PetscErrorCode PetscCLEAR(PetscObject* obj) noexcept nogil:
    if obj    == NULL: return PETSC_SUCCESS
    if obj[0] == NULL: return PETSC_SUCCESS
    cdef PetscObject tmp
    tmp = obj[0]; obj[0] = NULL
    return PetscObjectDestroy(&tmp)

cdef inline PetscErrorCode PetscDEALLOC(PetscObject* obj) noexcept nogil:
    if obj    == NULL: return PETSC_SUCCESS
    if obj[0] == NULL: return PETSC_SUCCESS
    cdef PetscObject tmp
    tmp = obj[0]; obj[0] = NULL
    if not (<int>PetscInitializeCalled): return PETSC_SUCCESS
    if     (<int>PetscFinalizeCalled):   return PETSC_SUCCESS
    return PetscObjectDelayedDestroy(&tmp)

cdef inline PetscErrorCode PetscINCSTATE(PetscObject *obj) noexcept nogil:
    if obj    == NULL: return PETSC_SUCCESS
    if obj[0] == NULL: return PETSC_SUCCESS
    return PetscObjectStateIncrease(obj[0])

# --------------------------------------------------------------------

cdef extern from *:
    ctypedef struct PyObject
    void _Py_DecRef"Py_DECREF"(PyObject*)
    PyObject* PyDict_New() except NULL
    PyObject* PyDict_GetItem(PyObject*, PyObject*) except *
    int       PyDict_SetItem(PyObject*, PyObject*, PyObject*) except -1
    int       PyDict_DelItem(PyObject*, PyObject*) except -1

cdef extern from * nogil:
    ctypedef struct _p_PetscObject:
        MPI_Comm comm
        const char *prefix
        PetscInt refct
        void *python_context
        PetscErrorCode (*python_destroy)(void*) noexcept nogil

cdef inline void Py_DecRef(PyObject *ob) noexcept with gil:
    _Py_DecRef(ob)

cdef PetscErrorCode PetscDelPyDict(void* ptr) noexcept nogil:
    if ptr != NULL and Py_IsInitialized():
        Py_DecRef(<PyObject*>ptr)
    return PETSC_SUCCESS

cdef object PetscGetPyDict(PetscObject obj, bint create):
    if obj.python_context != NULL:
        return <object>obj.python_context
    if create:
        obj.python_destroy = PetscDelPyDict
        obj.python_context = <void*>PyDict_New()
        return <object>obj.python_context
    return None

cdef inline object PetscGetPyObj(PetscObject o, char name[]):
    cdef object dct = PetscGetPyDict(o, False)
    if dct is None: return None
    cdef object key = bytes2str(name)
    cdef PyObject *d = <PyObject*>dct
    cdef PyObject *k = <PyObject*>key
    cdef PyObject *v = NULL
    v = PyDict_GetItem(d, k)
    if v != NULL: return <object>v
    return None

cdef inline object PetscSetPyObj(PetscObject o, char name[], object p):
    cdef object dct
    if p is not None:
        dct = PetscGetPyDict(o, True)
    else:
        dct = PetscGetPyDict(o, False)
        if dct is None: return None
    cdef str key = bytes2str(name)
    cdef PyObject *d = <PyObject*>dct
    cdef PyObject *k = <PyObject*>key
    cdef PyObject *v = <PyObject*>p
    PyDict_SetItem(d, k, v)
    if v == <PyObject*>None:
        PyDict_DelItem(d, k)
    return None

# --------------------------------------------------------------------

cdef extern from *:
    object PyLong_FromVoidPtr(void*)

cdef inline Py_intptr_t Object_toFortran(PetscObject o) nogil:
    return <Py_intptr_t> o

# --------------------------------------------------------------------

cdef inline type subtype_DM(PetscDM dm):
    cdef PetscObject obj = <PetscObject> dm
    if obj == NULL: return DM
    # ---
    cdef PetscBool match = PETSC_FALSE
    CHKERR(PetscObjectTypeCompare(obj, b"da", &match))
    if match == PETSC_TRUE: return DMDA
    CHKERR(PetscObjectTypeCompare(obj, b"plex", &match))
    if match == PETSC_TRUE: return DMPlex
    CHKERR(PetscObjectTypeCompare(obj, b"composite", &match))
    if match == PETSC_TRUE: return DMComposite
    CHKERR(PetscObjectTypeCompare(obj, b"shell", &match))
    if match == PETSC_TRUE: return DMShell
    CHKERR(PetscObjectTypeCompare(obj, b"stag", &match))
    if match == PETSC_TRUE: return DMStag
    CHKERR(PetscObjectTypeCompare(obj, b"swarm", &match))
    if match == PETSC_TRUE: return DMSwarm
    # ---
    return DM

cdef inline type subtype_Object(PetscObject obj):
    cdef type klass = Object
    if obj == NULL: return klass
    cdef PetscClassId classid = 0
    CHKERR(PetscObjectGetClassId(obj, &classid))
    if classid == PETSC_DM_CLASSID:
        klass = subtype_DM(<PetscDM>obj)
    else:
        klass = PyPetscType_Lookup(classid)
    return klass

# --------------------------------------------------------------------

cdef PetscErrorCode PetscObjectOptionsHandler_PYTHON(
    PetscObject obj,
    PetscOptionItems unused_PetscOptionsObject,
    void *unused_ctx,
   ) except PETSC_ERR_PYTHON with gil:
    cdef Object pobj = PyPetscObject_New(obj)
    cdef object handler = pobj.get_attr('__optshandler__')
    if handler is None: return PETSC_SUCCESS
    handler(pobj)
    return PETSC_SUCCESS



import os
import inspect
import textwrap


def is_cyfunction(obj):
    return type(obj).__name__ == 'cython_function_or_method'


def is_function(obj):
    return inspect.isbuiltin(obj) or is_cyfunction(obj) or type(obj) is type(ord)


def is_method(obj):
    return (
        inspect.ismethoddescriptor(obj)
        or inspect.ismethod(obj)
        or is_cyfunction(obj)
        or type(obj)
        in (
            type(str.index),
            type(str.__add__),
            type(str.__new__),
        )
    )


def is_classmethod(obj):
    return inspect.isbuiltin(obj) or type(obj).__name__ in (
        'classmethod',
        'classmethod_descriptor',
    )


def is_staticmethod(obj):
    return type(obj).__name__ in ('staticmethod',)


def is_constant(obj):
    return isinstance(obj, (int, float, str))


def is_datadescr(obj):
    return inspect.isdatadescriptor(obj) and not hasattr(obj, 'fget')


def is_property(obj):
    return inspect.isdatadescriptor(obj) and hasattr(obj, 'fget')


def is_class(obj):
    return inspect.isclass(obj) or type(obj) is type(int)


class Lines(list):
    INDENT = ' ' * 4
    level = 0

    @property
    def add(self):
        return self

    @add.setter
    def add(self, lines):
        if lines is None:
            return
        if isinstance(lines, str):
            lines = textwrap.dedent(lines).strip().split('\n')
        indent = self.INDENT * self.level
        for line in lines:
            self.append(indent + line)


def signature(obj):
    doc = obj.__doc__
    doc = doc or f'{obj.__name__}: Any'  # FIXME remove line
    sig = doc.split('\n', 1)[0].split('.', 1)[-1]
    return sig or None


def visit_constant(constant):
    name, value = constant
    return f'{name}: Final[{type(value).__name__}] = ...'


def visit_function(function):
    sig = signature(function)
    return f'def {sig}: ...'


def visit_method(method):
    sig = signature(method)
    return f'def {sig}: ...'


def visit_datadescr(datadescr):
    sig = signature(datadescr)
    return f'{sig}'


def visit_property(prop, name=None):
    sig = signature(prop.fget)
    pname = name or prop.fget.__name__
    ptype = sig.rsplit('->', 1)[-1].strip()
    return f'{pname}: {ptype}'


def visit_constructor(cls, name='__init__', args=None):
    init = name == '__init__'
    argname = cls.__name__.lower()
    argtype = cls.__name__
    initarg = args or f'{argname}: Optional[{argtype}] = None'
    selfarg = 'self' if init else 'cls'
    rettype = 'None' if init else argtype
    arglist = f'{selfarg}, {initarg}'
    sig = f'{name}({arglist}) -> {rettype}'
    return f'def {sig}: ...'


def visit_class(cls, outer=None, done=None):
    skip = {
        '__doc__',
        '__dict__',
        '__module__',
        '__weakref__',
        '__pyx_vtable__',
        '__enum2str',  # FIXME refactor implementation
        '_traceback_',  # FIXME maybe refactor?
        '__lt__',
        '__le__',
        '__ge__',
        '__gt__',
    }
    special = {
        '__len__': '__len__(self) -> int',
        '__bool__': '__bool__(self) -> bool',
        '__hash__': '__hash__(self) -> int',
        '__int__': '__int__(self) -> int',
        '__index__': '__int__(self) -> int',
        '__str__': '__str__(self) -> str',
        '__repr__': '__repr__(self) -> str',
        '__eq__': '__eq__(self, other: object) -> bool',
        '__ne__': '__ne__(self, other: object) -> bool',
    }
    constructor = (
        '__new__',
        '__init__',
    )

    qualname = cls.__name__
    cls_name = cls.__name__
    if outer is not None and cls_name.startswith(outer):
        cls_name = cls_name[len(outer) :]
        qualname = f'{outer}.{cls_name}'

    override = OVERRIDE.get(qualname, {})
    done = set() if done is None else done
    lines = Lines()

    try:

        class sub(cls):
            pass

        final = False
    except TypeError:
        final = True
    if final:
        lines.add = '@final'
    base = cls.__base__
    if base is object:
        lines.add = f'class {cls_name}:'
    else:
        lines.add = f'class {cls_name}({base.__name__}):'
    lines.level += 1
    start = len(lines)

    for name in constructor:
        if name in cls.__dict__:
            done.add(name)

    if '__hash__' in cls.__dict__:
        if cls.__hash__ is None:
            done.add('__hash__')

    dct = cls.__dict__
    keys = list(dct.keys())

    def dunder(name):
        return name.startswith('__') and name.endswith('__')

    def members(seq):
        for name in seq:
            if name in skip:
                continue
            if name in done:
                continue
            if dunder(name):
                if name not in special and name not in override:
                    done.add(name)
                    continue
            yield name

    for name in members(keys):
        attr = getattr(cls, name)
        if is_class(attr):
            done.add(name)
            lines.add = visit_class(attr, outer=cls_name)
            continue

    for name in members(keys):
        if name in override:
            done.add(name)
            lines.add = override[name]
            continue

        if name in special:
            done.add(name)
            sig = special[name]
            lines.add = f'def {sig}: ...'
            continue

        attr = getattr(cls, name)

        if is_method(attr):
            done.add(name)
            if name == attr.__name__:
                obj = dct[name]
                if is_classmethod(obj):
                    lines.add = '@classmethod'
                elif is_staticmethod(obj):
                    lines.add = '@staticmethod'
                lines.add = visit_method(attr)
            elif True:
                lines.add = f'{name} = {attr.__name__}'
            continue

        if is_datadescr(attr):
            done.add(name)
            lines.add = visit_datadescr(attr)
            continue

        if is_property(attr):
            done.add(name)
            lines.add = visit_property(attr, name)
            continue

        if is_constant(attr):
            done.add(name)
            lines.add = visit_constant((name, attr))
            continue

    leftovers = [name for name in keys if name not in done and name not in skip]
    if leftovers:
        raise RuntimeError(f'leftovers: {leftovers}')

    if len(lines) == start:
        lines.add = 'pass'
    lines.level -= 1
    return lines


def visit_module(module, done=None):
    skip = {
        '__doc__',
        '__name__',
        '__loader__',
        '__spec__',
        '__file__',
        '__package__',
        '__builtins__',
        '__pyx_unpickle_Enum',  # FIXME review
    }

    done = set() if done is None else done
    lines = Lines()

    keys = list(module.__dict__.keys())
    keys.sort(key=lambda name: name.startswith('_'))

    constants = [
        (name, getattr(module, name))
        for name in keys
        if all(
            (
                name not in done and name not in skip,
                isinstance(getattr(module, name), int),
            )
        )
    ]
    for name, value in constants:
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_constant((name, value))
    if constants:
        lines.add = ''

    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)

        if is_class(value):
            done.add(name)
            if value.__module__ != module.__name__:
                continue
            lines.add = visit_class(value)
            lines.add = ''
            instances = [
                (k, getattr(module, k))
                for k in keys
                if all(
                    (
                        k not in done and k not in skip,
                        type(getattr(module, k)) is value,
                    )
                )
            ]
            for attrname, attrvalue in instances:
                done.add(attrname)
                lines.add = visit_constant((attrname, attrvalue))
            if instances:
                lines.add = ''
            continue

        if is_function(value):
            done.add(name)
            if name == value.__name__:
                lines.add = visit_function(value)
            else:
                lines.add = f'{name} = {value.__name__}'
            continue

    lines.add = ''
    for name in keys:
        if name in done or name in skip:
            continue
        value = getattr(module, name)
        done.add(name)
        if name in OVERRIDE:
            lines.add = OVERRIDE[name]
        else:
            lines.add = visit_constant((name, value))

    leftovers = [name for name in keys if name not in done and name not in skip]
    if leftovers:
        raise RuntimeError(f'leftovers: {leftovers}')
    return lines


IMPORTS = """
from __future__ import annotations
import sys
from threading import Lock
from typing import (
    Any,
    Union,
    Optional,
    NoReturn,
    overload,
)
if sys.version_info >= (3, 8):
    from typing import (
        final,
        Final,
        Literal,
    )
else:
    from typing_extensions import (
        final,
        Final,
        Literal,
    )
if sys.version_info >= (3, 9):
    from collections.abc import (
        Callable,
        Hashable,
        Iterable,
        Iterator,
        Sequence,
        Mapping,
    )
else:
    from typing import (
        Callable,
        Hashable,
        Iterable,
        Iterator,
        Sequence,
        Mapping,
    )
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from os import PathLike

import numpy

IntType: numpy.dtype = ...
RealType: numpy.dtype = ...
ComplexType: numpy.dtype = ...
ScalarType: numpy.dtype = ...
"""

OVERRIDE = {
    'Error': {},
    '__pyx_capi__': '__pyx_capi__: Final[Dict[str, Any]] = ...',
    '__type_registry__': '__type_registry__: Final[Dict[int, type[Object]]] = ...',
}

TYPING = """
"""


def visit_petsc4py_PETSc(done=None):
    from petsc4py import PETSc as module

    lines = Lines()
    lines.add = IMPORTS
    lines.add = ''
    lines.add = visit_module(module)
    lines.add = TYPING
    return lines


def generate(filename):
    dirname = os.path.dirname(filename)
    os.makedirs(dirname, exist_ok=True)
    with open(filename, 'w') as f:
        for line in visit_petsc4py_PETSc():
            print(line, file=f)


OUTDIR = os.path.join('src', 'petsc4py')

if __name__ == '__main__':
    generate(os.path.join(OUTDIR, 'PETSc.pyi'))



from petsc4py import PETSc
import unittest

# --------------------------------------------------------------------


class BaseTestDA:
    COMM = PETSc.COMM_WORLD
    SIZES = None
    BOUNDARY = None
    DOF = 1
    STENCIL = PETSc.DMDA.StencilType.STAR
    SWIDTH = 1

    def setUp(self):
        self.da = PETSc.DMDA().create(
            dim=len(self.SIZES),
            dof=self.DOF,
            sizes=self.SIZES,
            boundary_type=self.BOUNDARY,
            stencil_type=self.STENCIL,
            stencil_width=self.SWIDTH,
            comm=self.COMM,
        )

    def tearDown(self):
        self.da = None
        PETSc.garbage_cleanup()

    def testGetInfo(self):
        dim = self.da.getDim()
        dof = self.da.getDof()
        sizes = self.da.getSizes()
        boundary = self.da.getBoundaryType()
        stencil_type = self.da.getStencilType()
        stencil_width = self.da.getStencilWidth()
        self.assertEqual(dim, len(self.SIZES))
        self.assertEqual(dof, self.DOF)
        self.assertEqual(sizes, tuple(self.SIZES))
        self.assertEqual(boundary, self.BOUNDARY or (0,) * dim)
        self.assertEqual(stencil_type, self.STENCIL)
        self.assertEqual(stencil_width, self.SWIDTH)

    def testRangesCorners(self):
        dim = self.da.getDim()
        ranges = self.da.getRanges()
        starts, lsizes = self.da.getCorners()
        self.assertEqual(dim, len(ranges))
        self.assertEqual(dim, len(starts))
        self.assertEqual(dim, len(lsizes))
        for i in range(dim):
            s, e = ranges[i]
            self.assertEqual(s, starts[i])
            self.assertEqual(e - s, lsizes[i])

    def testGhostRangesCorners(self):
        dim = self.da.getDim()
        ranges = self.da.getGhostRanges()
        starts, lsizes = self.da.getGhostCorners()
        self.assertEqual(dim, len(ranges))
        self.assertEqual(dim, len(starts))
        self.assertEqual(dim, len(lsizes))
        for i in range(dim):
            s, e = ranges[i]
            self.assertEqual(s, starts[i])
            self.assertEqual(e - s, lsizes[i])

    def testOwnershipRanges(self):
        ownership_ranges = self.da.getOwnershipRanges()
        procsizes = self.da.getProcSizes()
        self.assertEqual(len(procsizes), len(ownership_ranges))
        for i, m in enumerate(procsizes):
            self.assertEqual(m, len(ownership_ranges[i]))

    def testFieldName(self):
        for i in range(self.da.getDof()):
            self.da.setFieldName(i, 'field%d' % i)
        for i in range(self.da.getDof()):
            name = self.da.getFieldName(i)
            self.assertEqual(name, 'field%d' % i)

    def testCoordinates(self):
        self.da.setUniformCoordinates(0, 1, 0, 1, 0, 1)
        #
        c = self.da.getCoordinates()
        self.da.setCoordinates(c)
        c.destroy()
        cda = self.da.getCoordinateDM()
        cda.destroy()
        #
        c = self.da.getCoordinates()
        self.da.setCoordinates(c)
        c.destroy()
        gc = self.da.getCoordinatesLocal()
        gc.destroy()

    def testCreateVecMat(self):
        vn = self.da.createNaturalVec()
        vg = self.da.createGlobalVec()
        vl = self.da.createLocalVec()
        mat = self.da.createMat()
        self.assertTrue(mat.getType() in ('aij', 'seqaij', 'mpiaij'))
        vn.set(1.0)
        self.da.naturalToGlobal(vn, vg)
        self.assertEqual(vg.max()[1], 1.0)
        self.assertEqual(vg.min()[1], 1.0)
        self.da.globalToLocal(vg, vl)
        self.assertEqual(vl.max()[1], 1.0)
        self.assertTrue(vl.min()[1] in (1.0, 0.0))
        vn.set(0.0)
        self.da.globalToNatural(vg, vn)
        self.assertEqual(vn.max()[1], 1.0)
        self.assertEqual(vn.min()[1], 1.0)
        vl2 = self.da.createLocalVec()
        self.da.localToLocal(vl, vl2)
        self.assertEqual(vl2.max()[1], 1.0)
        self.assertTrue(vl2.min()[1] in (1.0, 0.0))
        NONE = PETSc.DM.BoundaryType.NONE
        btype = self.da.boundary_type
        psize = self.da.proc_sizes
        for b, p in zip(btype, psize):
            if b != NONE and p == 1:
                return
        vg2 = self.da.createGlobalVec()
        self.da.localToGlobal(vl2, vg2)

    def testGetVec(self):
        vg = self.da.getGlobalVec()
        vl = self.da.getLocalVec()
        try:
            vg.set(1.0)
            self.assertEqual(vg.max()[1], 1.0)
            self.assertEqual(vg.min()[1], 1.0)
            self.da.globalToLocal(vg, vl)
            self.assertEqual(vl.max()[1], 1.0)
            self.assertTrue(vl.min()[1] in (1.0, 0.0))
            vl.set(2.0)
            NONE = PETSc.DM.BoundaryType.NONE
            btype = self.da.boundary_type
            psize = self.da.proc_sizes
            for b, p in zip(btype, psize):
                if b != NONE and p == 1:
                    return
            self.da.localToGlobal(vl, vg)
            self.assertEqual(vg.max()[1], 2.0)
            self.assertTrue(vg.min()[1] in (2.0, 0.0))
        finally:
            self.da.restoreGlobalVec(vg)
            self.da.restoreLocalVec(vl)
            self.assertFalse(vg)
            self.assertFalse(vl)

        name = 'abcd'

        vg = self.da.getGlobalVec(name)
        vg.set(4.0)
        self.da.restoreGlobalVec(vg, name)
        self.assertFalse(vg)
        vg = self.da.getGlobalVec()
        vg.setRandom()
        self.da.restoreGlobalVec(vg)
        vg = self.da.getGlobalVec(name)
        vg.shift(-4.0)
        self.assertEqual(vg.max()[1], 0.0)
        self.da.restoreGlobalVec(vg, name)

        vl = self.da.getLocalVec(name)
        vl.set(4.0)
        self.da.restoreLocalVec(vl, name)
        self.assertFalse(vl)
        vl = self.da.getLocalVec()
        vl.setRandom()
        self.da.restoreLocalVec(vl)
        vl = self.da.getLocalVec(name)
        vl.shift(-4.0)
        self.assertEqual(vl.max()[1], 0.0)
        self.da.restoreLocalVec(vl, name)

    def testGetArray(self):
        for r in [True, False]:
            vg = self.da.getGlobalVec()
            vl = self.da.getLocalVec()

            ag = self.da.getVecArray(vg, readonly=r)
            al = self.da.getVecArray(vl, readonly=r)

            # test reading lower-left and upper-right corners
            ranges = self.da.getRanges()
            ranges = list(zip(*ranges))
            dim = len(ranges[0])
            if dim == 1:
                _ = ag[ranges[0][0] : ranges[1][0]]
                _ = al[ranges[0][0] : ranges[1][0]]
            elif dim == 2:
                _ = ag[ranges[0][0] : ranges[1][0], ranges[0][1] : ranges[1][1]]
                _ = al[ranges[0][0] : ranges[1][0], ranges[0][1] : ranges[1][1]]
            elif dim == 3:
                _ = ag[
                    ranges[0][0] : ranges[1][0],
                    ranges[0][1] : ranges[1][1],
                    ranges[0][2] : ranges[1][2],
                ]
                _ = al[
                    ranges[0][0] : ranges[1][0],
                    ranges[0][1] : ranges[1][1],
                    ranges[0][2] : ranges[1][2],
                ]

            # test writing
            if not r:
                ag[:] = 1.0
                al[:] = 1.0
            else:
                with self.assertRaises(ValueError):
                    ag[:] = 1.0
                with self.assertRaises(ValueError):
                    al[:] = 1.0

            self.da.restoreGlobalVec(vg)
            self.da.restoreLocalVec(vl)

    def testGetOther(self):
        _ = self.da.getAO()
        _ = self.da.getLGMap()
        _, _ = self.da.getScatter()

    def testRefineCoarsen(self):
        da = self.da
        rda = da.refine()
        self.assertEqual(da.getDim(), rda.getDim())
        self.assertEqual(da.getDof(), rda.getDof())
        if da.dim != 1:
            self.assertEqual(da.getStencilType(), rda.getStencilType())
        self.assertEqual(da.getStencilWidth(), rda.getStencilWidth())
        cda = rda.coarsen()
        self.assertEqual(rda.getDim(), cda.getDim())
        self.assertEqual(rda.getDof(), cda.getDof())
        for n1, n2 in zip(self.da.getSizes(), cda.getSizes()):
            self.assertTrue(abs(n1 - n2) <= 1)

    def testCoarsenRefine(self):
        if PETSc.COMM_WORLD.getSize() > 6:
            return
        da = self.da
        cda = self.da.coarsen()
        self.assertEqual(da.getDim(), cda.getDim())
        self.assertEqual(da.getDof(), cda.getDof())
        if da.dim != 1:
            self.assertEqual(da.getStencilType(), cda.getStencilType())
        self.assertEqual(da.getStencilWidth(), cda.getStencilWidth())
        rda = cda.refine()
        for n1, n2 in zip(self.da.getSizes(), rda.getSizes()):
            self.assertTrue(abs(n1 - n2) <= 1)

    def testRefineHierarchy(self):
        levels = self.da.refineHierarchy(2)
        self.assertTrue(isinstance(levels, list))
        self.assertEqual(len(levels), 2)
        for item in levels:
            self.assertTrue(isinstance(item, PETSc.DM))

    def testCoarsenHierarchy(self):
        if PETSc.COMM_WORLD.getSize() > 6:
            return
        levels = self.da.coarsenHierarchy(2)
        self.assertTrue(isinstance(levels, list))
        self.assertEqual(len(levels), 2)
        for item in levels:
            self.assertTrue(isinstance(item, PETSc.DM))

    def testCreateInterpolation(self):
        da = self.da
        if da.dim == 1:
            return
        rda = da.refine()
        _, _ = da.createInterpolation(rda)

    def testCreateInjection(self):
        if PETSc.COMM_WORLD.getSize() > 6:
            return
        da = self.da
        if da.dim == 1:
            return
        rda = da.refine()
        _ = da.createInjection(rda)

    def testzeroRowsColumnsStencil(self):
        da = self.da
        A = da.createMatrix()
        x = da.createGlobalVector()
        x.set(2.0)
        A.setDiagonal(x)
        diag1 = x.duplicate()
        A.getDiagonal(diag1)
        if self.SIZES != 2:  # only coded test for 2D case
            return
        istart, iend, jstart, jend = da.getRanges()
        self.assertTrue(x.equal(diag1))
        zeroidx = []
        for i in range(istart, iend):
            for j in range(jstart, jend):
                row = PETSc.Mat.Stencil()
                row.index = (i, j)
                zeroidx = zeroidx + [row]
        diag2 = x.duplicate()
        diag2.set(1.0)
        A.zeroRowsColumnsStencil(zeroidx, 1.0, x, diag2)
        ans = x.duplicate()
        ans.set(2.0)
        self.assertTrue(ans.equal(diag2))


MIRROR = PETSc.DMDA.BoundaryType.MIRROR
GHOSTED = PETSc.DMDA.BoundaryType.GHOSTED
PERIODIC = PETSc.DMDA.BoundaryType.PERIODIC
TWIST = PETSc.DMDA.BoundaryType.TWIST

SCALE = 4


class BaseTestDA_1D(BaseTestDA):
    SIZES = [100 * SCALE]


class BaseTestDA_2D(BaseTestDA):
    SIZES = [9 * SCALE, 11 * SCALE]


class BaseTestDA_3D(BaseTestDA):
    SIZES = [6 * SCALE, 7 * SCALE, 8 * SCALE]


# --------------------------------------------------------------------


class TestDA_1D(BaseTestDA_1D, unittest.TestCase):
    pass


class TestDA_1D_W0(TestDA_1D):
    SWIDTH = 0


class TestDA_1D_W2(TestDA_1D):
    SWIDTH = 2


class TestDA_2D(BaseTestDA_2D, unittest.TestCase):
    pass


class TestDA_2D_W0(TestDA_2D):
    SWIDTH = 0


class TestDA_2D_W0_N2(TestDA_2D):
    DOF = 2
    SWIDTH = 0


class TestDA_2D_W2(TestDA_2D):
    SWIDTH = 2


class TestDA_2D_W2_N2(TestDA_2D):
    DOF = 2
    SWIDTH = 2


class TestDA_2D_PXY(TestDA_2D):
    SIZES = [13 * SCALE, 17 * SCALE]
    DOF = 2
    SWIDTH = 5
    BOUNDARY = (PERIODIC,) * 2


class TestDA_2D_GXY(TestDA_2D):
    SIZES = [13 * SCALE, 17 * SCALE]
    DOF = 2
    SWIDTH = 5
    BOUNDARY = (GHOSTED,) * 2


class TestDA_2D_TXY(TestDA_2D):
    SIZES = [13 * SCALE, 17 * SCALE]
    DOF = 2
    SWIDTH = 5
    BOUNDARY = (TWIST,) * 2


class TestDA_3D(BaseTestDA_3D, unittest.TestCase):
    pass


class TestDA_3D_W0(TestDA_3D):
    SWIDTH = 0


class TestDA_3D_W0_N2(TestDA_3D):
    DOF = 2
    SWIDTH = 0


class TestDA_3D_W2(TestDA_3D):
    SWIDTH = 2


class TestDA_3D_W2_N2(TestDA_3D):
    DOF = 2
    SWIDTH = 2


class TestDA_3D_PXYZ(TestDA_3D):
    SIZES = [11 * SCALE, 13 * SCALE, 17 * SCALE]
    DOF = 2
    SWIDTH = 3
    BOUNDARY = (PERIODIC,) * 3


class TestDA_3D_GXYZ(TestDA_3D):
    SIZES = [11 * SCALE, 13 * SCALE, 17 * SCALE]
    DOF = 2
    SWIDTH = 3
    BOUNDARY = (GHOSTED,) * 3


class TestDA_3D_TXYZ(TestDA_3D):
    SIZES = [11 * SCALE, 13 * SCALE, 17 * SCALE]
    DOF = 2
    SWIDTH = 3
    BOUNDARY = (TWIST,) * 3


# --------------------------------------------------------------------

DIM = (
    1,
    2,
    3,
)
DOF = (
    None,
    1,
    2,
    3,
    4,
    5,
)
BOUNDARY_TYPE = (
    None,
    'none',
    (0,) * 3,
    0,
    'ghosted',
    (GHOSTED,) * 3,
    GHOSTED,
    'periodic',
    (PERIODIC,) * 3,
    PERIODIC,
    'twist',
    (TWIST,) * 3,
    TWIST,
)
STENCIL_TYPE = (None, 'star', 'box')
STENCIL_WIDTH = (None, 0, 1, 2, 3)


DIM = (1, 2, 3)
DOF = (None, 2, 5)
BOUNDARY_TYPE = (None, 'none', 'periodic', 'ghosted', 'twist')
STENCIL_TYPE = (None, 'box')
STENCIL_WIDTH = (None, 1, 2)


class TestDACreate(unittest.TestCase):
    pass


counter = 0
for dim in DIM:
    for dof in DOF:
        for boundary in BOUNDARY_TYPE:
            if isinstance(boundary, tuple):
                boundary = boundary[:dim]
            for stencil in STENCIL_TYPE:
                for width in STENCIL_WIDTH:
                    kargs = {
                        'sizes': [8 * SCALE] * dim,
                        'dim': dim,
                        'dof': dof,
                        'boundary_type': boundary,
                        'stencil_type': stencil,
                        'stencil_width': width,
                    }

                    def testCreate(self, kargs=kargs):
                        kargs = dict(kargs)
                        da = PETSc.DMDA().create(**kargs)
                        da.destroy()

                    setattr(TestDACreate, 'testCreate%04d' % counter, testCreate)
                    del testCreate, kargs
                    counter += 1
del counter, dim, dof, boundary, stencil, width


class TestDADuplicate(unittest.TestCase):
    pass


counter = 0
for dim in DIM:
    for dof in DOF:
        for boundary in BOUNDARY_TYPE:
            if isinstance(boundary, tuple):
                boundary = boundary[:dim]
            for stencil in STENCIL_TYPE:
                for width in STENCIL_WIDTH:
                    kargs = {
                        'dim': dim,
                        'dof': dof,
                        'boundary_type': boundary,
                        'stencil_type': stencil,
                        'stencil_width': width,
                    }

                    def testDuplicate(self, kargs=kargs):
                        kargs = dict(kargs)
                        dim = kargs.pop('dim')
                        dof = kargs['dof']
                        boundary = kargs['boundary_type']
                        stencil = kargs['stencil_type']
                        width = kargs['stencil_width']
                        da = PETSc.DMDA().create([8 * SCALE] * dim)
                        newda = da.duplicate(**kargs)
                        self.assertEqual(newda.dim, da.dim)
                        self.assertEqual(newda.sizes, da.sizes)
                        self.assertEqual(newda.proc_sizes, da.proc_sizes)
                        self.assertEqual(newda.ranges, da.ranges)
                        self.assertEqual(newda.corners, da.corners)
                        if (
                            newda.boundary_type == da.boundary_type
                            and newda.stencil_width == da.stencil_width
                        ):
                            self.assertEqual(newda.ghost_ranges, da.ghost_ranges)
                            self.assertEqual(newda.ghost_corners, da.ghost_corners)
                        if dof is None:
                            dof = da.dof
                        if boundary is None:
                            boundary = da.boundary_type
                        elif boundary == 'none':
                            boundary = (0,) * dim
                        elif boundary == 'mirror':
                            boundary = (MIRROR,) * dim
                        elif boundary == 'ghosted':
                            boundary = (GHOSTED,) * dim
                        elif boundary == 'periodic':
                            boundary = (PERIODIC,) * dim
                        elif boundary == 'twist':
                            boundary = (TWIST,) * dim
                        elif isinstance(boundary, int):
                            boundary = (boundary,) * dim
                        if stencil is None:
                            stencil = da.stencil[0]
                        if width is None:
                            width = da.stencil_width
                        self.assertEqual(newda.dof, dof)
                        self.assertEqual(newda.boundary_type, boundary)
                        if dim == 1:
                            self.assertEqual(newda.stencil, (stencil, width))
                        newda.destroy()
                        da.destroy()

                    setattr(
                        TestDADuplicate, 'testDuplicate%04d' % counter, testDuplicate
                    )
                    del testDuplicate, kargs
                    counter += 1
del counter, dim, dof, boundary, stencil, width

# --------------------------------------------------------------------

if PETSc.COMM_WORLD.getSize() > 1:
    del TestDA_1D_W0
    del TestDA_2D_W0, TestDA_2D_W0_N2
    del TestDA_3D_W0, TestDA_3D_W0_N2

# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



# --------------------------------------------------------------------

class DMType(object):
    """`DM` types."""
    DA        = S_(DMDA_type)
    COMPOSITE = S_(DMCOMPOSITE)
    SLICED    = S_(DMSLICED)
    SHELL     = S_(DMSHELL)
    PLEX      = S_(DMPLEX)
    REDUNDANT = S_(DMREDUNDANT)
    PATCH     = S_(DMPATCH)
    MOAB      = S_(DMMOAB)
    NETWORK   = S_(DMNETWORK)
    FOREST    = S_(DMFOREST)
    P4EST     = S_(DMP4EST)
    P8EST     = S_(DMP8EST)
    SWARM     = S_(DMSWARM)
    PRODUCT   = S_(DMPRODUCT)
    STAG      = S_(DMSTAG)


class DMBoundaryType(object):
    """`DM` Boundary types."""
    NONE     = DM_BOUNDARY_NONE
    GHOSTED  = DM_BOUNDARY_GHOSTED
    MIRROR   = DM_BOUNDARY_MIRROR
    PERIODIC = DM_BOUNDARY_PERIODIC
    TWIST    = DM_BOUNDARY_TWIST


class DMPolytopeType(object):
    """The `DM` cell types."""
    POINT              = DM_POLYTOPE_POINT
    SEGMENT            = DM_POLYTOPE_SEGMENT
    POINT_PRISM_TENSOR = DM_POLYTOPE_POINT_PRISM_TENSOR
    TRIANGLE           = DM_POLYTOPE_TRIANGLE
    QUADRILATERAL      = DM_POLYTOPE_QUADRILATERAL
    SEG_PRISM_TENSOR   = DM_POLYTOPE_SEG_PRISM_TENSOR
    TETRAHEDRON        = DM_POLYTOPE_TETRAHEDRON
    HEXAHEDRON         = DM_POLYTOPE_HEXAHEDRON
    TRI_PRISM          = DM_POLYTOPE_TRI_PRISM
    TRI_PRISM_TENSOR   = DM_POLYTOPE_TRI_PRISM_TENSOR
    QUAD_PRISM_TENSOR  = DM_POLYTOPE_QUAD_PRISM_TENSOR
    PYRAMID            = DM_POLYTOPE_PYRAMID
    FV_GHOST           = DM_POLYTOPE_FV_GHOST
    INTERIOR_GHOST     = DM_POLYTOPE_INTERIOR_GHOST
    UNKNOWN            = DM_POLYTOPE_UNKNOWN
    UNKNOWN_CELL       = DM_POLYTOPE_UNKNOWN_CELL
    UNKNOWN_FACE       = DM_POLYTOPE_UNKNOWN_FACE


class DMReorderDefaultFlag(object):
    """The `DM` reordering default flags."""
    NOTSET = DM_REORDER_DEFAULT_NOTSET
    FALSE  = DM_REORDER_DEFAULT_FALSE
    TRUE   = DM_REORDER_DEFAULT_TRUE

# --------------------------------------------------------------------


cdef class DM(Object):
    """An object describing a computational grid or mesh."""

    Type         = DMType
    BoundaryType = DMBoundaryType
    PolytopeType = DMPolytopeType

    ReorderDefaultFlag = DMReorderDefaultFlag

    #

    def __cinit__(self):
        self.obj = <PetscObject*> &self.dm
        self.dm  = NULL

    def view(self, Viewer viewer=None) -> None:
        """View the `DM`.

        Collective.

        Parameters
        ----------
        viewer
            The DM viewer.

        See Also
        --------
        petsc.DMView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(DMView(self.dm, vwr))

    def load(self, Viewer viewer) -> Self:
        """Return a `DM` stored in binary.

        Collective.

        Parameters
        ----------
        viewer
            Viewer used to store the `DM`,
            like `Viewer.Type.BINARY` or `Viewer.Type.HDF5`.

        Notes
        -----
        When using `Viewer.Type.HDF5` format, one can save multiple `DMPlex` meshes
        in a single HDF5 files. This in turn requires one to name the `DMPlex`
        object with `Object.setName` before saving it with `DM.view` and before
        loading it with `DM.load` for identification of the mesh object.

        See Also
        --------
        DM.view, DM.load, Object.setName, petsc.DMLoad

        """
        CHKERR(DMLoad(self.dm, viewer.vwr))
        return self

    def destroy(self) -> Self:
        """Destroy the object.

        Collective.

        See Also
        --------
        petsc.DMDestroy

        """
        CHKERR(DMDestroy(&self.dm))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Return an empty `DM`.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.DMCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscDM newdm = NULL
        CHKERR(DMCreate(ccomm, &newdm))
        CHKERR(PetscCLEAR(self.obj)); self.dm = newdm
        return self

    def clone(self) -> DM:
        """Return the cloned `DM` .

        Collective.

        See Also
        --------
        petsc.DMClone

        """
        cdef DM dm = type(self)()
        CHKERR(DMClone(self.dm, &dm.dm))
        return dm

    def setType(self, dm_type: DM.Type | str) -> None:
        """Build a `DM`.

        Collective.

        Parameters
        ----------
        dm_type
            The type of `DM`.

        Notes
        -----
        `DM` types are available in `DM.Type` class.

        See Also
        --------
        DM.Type, petsc.DMSetType

        """
        cdef PetscDMType cval = NULL
        dm_type = str2bytes(dm_type, &cval)
        CHKERR(DMSetType(self.dm, cval))

    def getType(self) -> str:
        """Return the `DM` type name.

        Not collective.

        See Also
        --------
        petsc.DMGetType

        """
        cdef PetscDMType cval = NULL
        CHKERR(DMGetType(self.dm, &cval))
        return bytes2str(cval)

    def getDimension(self) -> int:
        """Return the topological dimension of the `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetDimension

        """
        cdef PetscInt dim = 0
        CHKERR(DMGetDimension(self.dm, &dim))
        return toInt(dim)

    def setDimension(self, dim: int) -> None:
        """Set the topological dimension of the `DM`.

        Collective.

        Parameters
        ----------
        dim
            Topological dimension.

        See Also
        --------
        petsc.DMSetDimension

        """
        cdef PetscInt cdim = asInt(dim)
        CHKERR(DMSetDimension(self.dm, cdim))

    def getCoordinateDim(self) -> int:
        """Return the dimension of embedding space for coordinates values.

        Not collective.

        See Also
        --------
        petsc.DMGetCoordinateDim

        """
        cdef PetscInt dim = 0
        CHKERR(DMGetCoordinateDim(self.dm, &dim))
        return toInt(dim)

    def setCoordinateDim(self, dim: int) -> None:
        """Set the dimension of embedding space for coordinates values.

        Not collective.

        Parameters
        ----------
        dim
            The embedding dimension.

        See Also
        --------
        petsc.DMSetCoordinateDim

        """
        cdef PetscInt cdim = asInt(dim)
        CHKERR(DMSetCoordinateDim(self.dm, cdim))

    def setOptionsPrefix(self, prefix: str | None) -> None:
        """Set the prefix used for searching for options in the database.

        Logically collective.

        See Also
        --------
        petsc_options, getOptionsPrefix, petsc.DMSetOptionsPrefix

        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR(DMSetOptionsPrefix(self.dm, cval))

    def getOptionsPrefix(self) -> str:
        """Return the prefix used for searching for options in the database.

        Not collective.

        See Also
        --------
        petsc_options, setOptionsPrefix, petsc.DMGetOptionsPrefix

        """
        cdef const char *cval = NULL
        CHKERR(DMGetOptionsPrefix(self.dm, &cval))
        return bytes2str(cval)

    def appendOptionsPrefix(self, prefix: str | None) -> None:
        """Append to the prefix used for searching for options in the database.

        Logically collective.

        See Also
        --------
        petsc_options, setOptionsPrefix, petsc.DMAppendOptionsPrefix

        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR(DMAppendOptionsPrefix(self.dm, cval))

    def setFromOptions(self) -> None:
        """Configure the object from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.DMSetFromOptions

        """
        CHKERR(DMSetFromOptions(self.dm))

    def setUp(self) -> Self:
        """Return the data structure.

        Collective.

        See Also
        --------
        petsc.DMSetUp

        """
        CHKERR(DMSetUp(self.dm))
        return self

    # --- application context ---

    def setAppCtx(self, appctx: Any) -> None:
        """Set the application context."""
        self.set_attr('__appctx__', appctx)

    def getAppCtx(self) -> Any:
        """Return the application context."""
        return self.get_attr('__appctx__')

    #

    def getUseNatural(self) -> bool:
        """Get the flag for constructing a global-to-natural map.

        Not collective.

        Returns
        -------
        useNatural : bool
            Whether a global-to-natural map is created.

        See Also
        --------
        petsc.DMGetUseNatural

        """
        cdef PetscBool uN  = PETSC_FALSE
        CHKERR(DMGetUseNatural(self.dm, &uN))
        return toBool(uN)

    def setUseNatural(self, useNatural : bool) -> None:
        """Set the flag for constructing a global-to-natural map.

        Not collective.

        Parameters
        ----------
        useNatural : bool
            Whether a global-to-natural map is created.

        See Also
        --------
        petsc.DMSetUseNatural

        """
        cdef PetscBool uN  = useNatural
        CHKERR(DMSetUseNatural(self.dm, uN))

    def setBasicAdjacency(self, useCone: bool, useClosure: bool) -> None:
        """Set the flags for determining variable influence.

        Not collective.

        Parameters
        ----------
        useCone : bool
            Whether adjacency uses cone information.
        useClosure : bool
            Whether adjacency is computed using full closure information.

        See Also
        --------
        petsc.DMSetBasicAdjacency

        """
        cdef PetscBool uC  = useCone
        cdef PetscBool uCl = useClosure
        CHKERR(DMSetBasicAdjacency(self.dm, uC, uCl))

    def getBasicAdjacency(self) -> tuple[bool, bool]:
        """Return the flags for determining variable influence.

        Not collective.

        Returns
        -------
        useCone : bool
            Whether adjacency uses cone information.
        useClosure : bool
            Whether adjacency is computed using full closure information.

        See Also
        --------
        petsc.DMGetBasicAdjacency

        """
        cdef PetscBool uC  = PETSC_FALSE
        cdef PetscBool uCl = PETSC_FALSE
        CHKERR(DMGetBasicAdjacency(self.dm, &uC, &uCl))
        return toBool(uC), toBool(uCl)

    def setFieldAdjacency(self, field: int, useCone: bool, useClosure: bool) -> None:
        """Set the flags for determining variable influence.

        Not collective.

        Parameters
        ----------
        field : int
            The field number.
        useCone : bool
            Whether adjacency uses cone information.
        useClosure : bool
            Whether adjacency is computed using full closure information.

        See Also
        --------
        petsc.DMSetAdjacency

        """
        cdef PetscInt  f   = asInt(field)
        cdef PetscBool uC  = useCone
        cdef PetscBool uCl = useClosure
        CHKERR(DMSetAdjacency(self.dm, f, uC, uCl))

    def getFieldAdjacency(self, field: int) -> tuple[bool, bool]:
        """Return the flags for determining variable influence.

        Not collective.

        Parameters
        ----------
        field
            The field number.

        Returns
        -------
        useCone : bool
            Whether adjacency uses cone information.
        useClosure : bool
            Whether adjacency is computed using full closure information.

        See Also
        --------
        petsc.DMGetAdjacency

        """
        cdef PetscInt  f   = asInt(field)
        cdef PetscBool uC  = PETSC_FALSE
        cdef PetscBool uCl = PETSC_FALSE
        CHKERR(DMGetAdjacency(self.dm, f, &uC, &uCl))
        return toBool(uC), toBool(uCl)

    #

    def createSubDM(self, fields: Sequence[int]) -> tuple[IS, DM]:
        """Return `IS` and `DM` encapsulating a subproblem.

        Not collective.

        Returns
        -------
        iset : IS
            The global indices for all the degrees of freedom.
        subdm : DM
            The `DM` for the subproblem.

        See Also
        --------
        petsc.DMCreateSubDM

        """
        cdef IS iset = IS()
        cdef DM subdm = DM()
        cdef PetscInt *ifields = NULL
        cdef PetscInt numFields = 0
        fields = iarray_i(fields, &numFields, &ifields)
        CHKERR(DMCreateSubDM(self.dm, numFields, ifields, &iset.iset, &subdm.dm))
        return iset, subdm

    #

    def setAuxiliaryVec(self, Vec aux, label: DMLabel | None, value=0, part=0) -> None:
        """Set an auxiliary vector for a specific region.

        Not collective.

        Parameters
        ----------
        aux
            The auxiliary vector.
        label
            The name of the `DMLabel`.
        value
            Indicate the region.
        part
            The equation part, or 0 is unused.

        See Also
        --------
        petsc.DMGetLabel, petsc.DMSetAuxiliaryVec

        """
        cdef PetscInt cvalue = asInt(value)
        cdef PetscInt cpart = asInt(part)
        cdef const char *cval = NULL
        cdef PetscDMLabel clbl = NULL
        label = str2bytes(label, &cval)
        if cval == NULL: cval = b"" # XXX Should be fixed upstream
        CHKERR(DMGetLabel(self.dm, cval, &clbl))
        CHKERR(DMSetAuxiliaryVec(self.dm, clbl, cvalue, cpart, aux.vec))

    def getAuxiliaryVec(self, label: str | None = None, value: int | None = 0, part: int | None = 0) -> Vec:
        """Return an auxiliary vector for region.

        Not collective.

        Parameters
        ----------
        label
            The name of the `DMLabel`.
        value
            Indicate the region.
        part
            The equation part, or 0 is unused.

        See Also
        --------
        DM.getLabel, petsc.DMGetAuxiliaryVec

        """
        cdef PetscInt cvalue = asInt(value)
        cdef PetscInt cpart = asInt(part)
        cdef const char *cval = NULL
        cdef PetscDMLabel clbl = NULL
        cdef Vec aux = Vec()
        label = str2bytes(label, &cval)
        if cval == NULL: cval = b"" # XXX Should be fixed upstream
        CHKERR(DMGetLabel(self.dm, cval, &clbl))
        CHKERR(DMGetAuxiliaryVec(self.dm, clbl, cvalue, cpart, &aux.vec))
        return aux

    def setNumFields(self, numFields: int) -> None:
        """Set the number of fields in the `DM`.

        Logically collective.

        See Also
        --------
        petsc.DMSetNumFields

        """
        cdef PetscInt cnum = asInt(numFields)
        CHKERR(DMSetNumFields(self.dm, cnum))

    def getNumFields(self) -> int:
        """Return the number of fields in the `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetNumFields

        """
        cdef PetscInt cnum = 0
        CHKERR(DMGetNumFields(self.dm, &cnum))
        return toInt(cnum)

    def setField(self, index: int, Object field, label: str | None = None) -> None:
        """Set the discretization object for a given `DM` field.

        Logically collective.

        Parameters
        ----------
        index
            The field number.
        field
            The discretization object.
        label
            The name of the label indicating the support of the field,
            or `None` for the entire mesh.

        See Also
        --------
        petsc.DMSetField

        """
        cdef PetscInt     cidx = asInt(index)
        cdef PetscObject  cobj = field.obj[0]
        cdef PetscDMLabel clbl = NULL
        assert label is None
        CHKERR(DMSetField(self.dm, cidx, clbl, cobj))

    def getField(self, index: int) -> tuple[Object, None]:
        """Return the discretization object for a given `DM` field.

        Not collective.

        Parameters
        ----------
        index
            The field number.

        See Also
        --------
        petsc.DMGetField

        """
        cdef PetscInt     cidx = asInt(index)
        cdef PetscObject  cobj = NULL
        cdef PetscDMLabel clbl = NULL
        CHKERR(DMGetField(self.dm, cidx, &clbl, &cobj))
        assert clbl == NULL
        cdef Object field = subtype_Object(cobj)()
        field.obj[0] = cobj
        CHKERR(PetscINCREF(field.obj))
        return (field, None) # TODO REVIEW

    def addField(self, Object field, label: str | None = None) -> None:
        """Add a field to a `DM` object.

        Logically collective.

        Parameters
        ----------
        field
            The discretization object.
        label
            The name of the label indicating the support of the field,
            or `None` for the entire mesh.

        See Also
        --------
        petsc.DMAddField

        """
        cdef PetscObject  cobj = field.obj[0]
        cdef PetscDMLabel clbl = NULL
        assert label is None
        CHKERR(DMAddField(self.dm, clbl, cobj))

    def clearFields(self) -> None:
        """Remove all fields from the `DM`.

        Logically collective.

        See Also
        --------
        petsc.DMClearFields

        """
        CHKERR(DMClearFields(self.dm))

    def copyFields(self, DM dm, minDegree = None, maxDegree = None) -> None:
        """Copy the discretizations of this `DM` into another `DM`.

        Collective.

        Parameters
        ----------
        dm
            The `DM` that the fields are copied into.
        minDegree
            The minimum polynommial degree for the discretization,
            or `None` for no limit
        maxDegree
            The maximum polynommial degree for the discretization,
            or `None` for no limit

        See Also
        --------
        petsc.DMCopyFields

        """
        cdef PetscInt mindeg = PETSC_DETERMINE
        if minDegree is None:
            pass
        else:
            mindeg = asInt(minDegree)
        cdef PetscInt maxdeg = PETSC_DETERMINE
        if maxDegree is None:
            pass
        else:
            maxdeg = asInt(maxDegree)
        CHKERR(DMCopyFields(self.dm, mindeg, maxdeg, dm.dm))

    def createDS(self) -> None:
        """Create discrete systems.

        Collective.

        See Also
        --------
        petsc.DMCreateDS

        """
        CHKERR(DMCreateDS(self.dm))

    def clearDS(self) -> None:
        """Remove all discrete systems from the `DM`.

        Logically collective.

        See Also
        --------
        petsc.DMClearDS

        """
        CHKERR(DMClearDS(self.dm))

    def getDS(self) -> DS:
        """Return default `DS`.

        Not collective.

        See Also
        --------
        petsc.DMGetDS

        """
        cdef DS ds = DS()
        CHKERR(DMGetDS(self.dm, &ds.ds))
        CHKERR(PetscINCREF(ds.obj))
        return ds

    def copyDS(self, DM dm, minDegree = None, maxDegree = None) -> None:
        """Copy the discrete systems for this `DM` into another `DM`.

        Collective.

        Parameters
        ----------
        dm
            The `DM` that the discrete fields are copied into.
        minDegree
            The minimum polynommial degree for the discretization,
            or `None` for no limit
        maxDegree
            The maximum polynommial degree for the discretization,
            or `None` for no limit

        See Also
        --------
        petsc.DMCopyDS

        """
        cdef PetscInt mindeg = PETSC_DETERMINE
        if minDegree is None:
            pass
        else:
            mindeg = asInt(minDegree)
        cdef PetscInt maxdeg = PETSC_DETERMINE
        if maxDegree is None:
            pass
        else:
            maxdeg = asInt(maxDegree)
        CHKERR(DMCopyDS(self.dm, mindeg, maxdeg, dm.dm))

    def copyDisc(self, DM dm) -> None:
        """Copy fields and discrete systems of a `DM` into another `DM`.

        Collective.

        Parameters
        ----------
        dm
            The `DM` that the fields and discrete systems are copied into.

        See Also
        --------
        petsc.DMCopyDisc

        """
        CHKERR(DMCopyDisc(self.dm, dm.dm))

    #

    def getBlockSize(self) -> int:
        """Return the inherent block size associated with a `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetBlockSize

        """
        cdef PetscInt bs = 1
        CHKERR(DMGetBlockSize(self.dm, &bs))
        return toInt(bs)

    def setVecType(self, vec_type: Vec.Type | str) -> None:
        """Set the type of vector.

        Logically collective.

        See Also
        --------
        Vec.Type, petsc.DMSetVecType

        """
        cdef PetscVecType vtype = NULL
        vec_type = str2bytes(vec_type, &vtype)
        CHKERR(DMSetVecType(self.dm, vtype))

    def createGlobalVec(self) -> Vec:
        """Return a global vector.

        Collective.

        See Also
        --------
        petsc.DMCreateGlobalVector

        """
        cdef Vec vg = Vec()
        CHKERR(DMCreateGlobalVector(self.dm, &vg.vec))
        return vg

    def createLocalVec(self) -> Vec:
        """Return a local vector.

        Not collective.

        See Also
        --------
        petsc.DMCreateLocalVector

        """
        cdef Vec vl = Vec()
        CHKERR(DMCreateLocalVector(self.dm, &vl.vec))
        return vl

    def getGlobalVec(self, name : str | None = None) -> Vec:
        """Return a global vector.

        Collective.

        Parameters
        ----------
        name
            The optional name to retrieve a persistent vector.

        Notes
        -----
        When done with the vector, it must be restored using `restoreGlobalVec`.

        See Also
        --------
        restoreGlobalVec, petsc.DMGetGlobalVector, petsc.DMGetNamedGlobalVector

        """
        cdef Vec vg = Vec()
        cdef const char *cname = NULL
        str2bytes(name, &cname)
        if cname != NULL:
            CHKERR(DMGetNamedGlobalVector(self.dm, cname, &vg.vec))
        else:
            CHKERR(DMGetGlobalVector(self.dm, &vg.vec))
        CHKERR(PetscINCREF(vg.obj))
        return vg

    def restoreGlobalVec(self, Vec vg, name : str | None = None) -> None:
        """Restore a global vector obtained with `getGlobalVec`.

        Logically collective.

        Parameters
        ----------
        vg
            The global vector.
        name
            The name used to retrieve the persistent vector, if any.

        See Also
        --------
        getGlobalVec, petsc.DMRestoreGlobalVector, petsc.DMRestoreNamedGlobalVector

        """
        cdef const char *cname = NULL
        str2bytes(name, &cname)
        CHKERR(PetscDECREF(vg.obj))
        if cname != NULL:
            CHKERR(DMRestoreNamedGlobalVector(self.dm, cname, &vg.vec))
        else:
            CHKERR(DMRestoreGlobalVector(self.dm, &vg.vec))

    def getLocalVec(self, name : str | None = None) -> Vec:
        """Return a local vector.

        Not collective.

        Parameters
        ----------
        name
            The optional name to retrieve a persistent vector.

        Notes
        -----
        When done with the vector, it must be restored using `restoreLocalVec`.

        See Also
        --------
        restoreLocalVec, petsc.DMGetLocalVector, petsc.DMGetNamedLocalVector

        """
        cdef Vec vl = Vec()
        cdef const char *cname = NULL
        str2bytes(name, &cname)
        if cname != NULL:
            CHKERR(DMGetNamedLocalVector(self.dm, cname, &vl.vec))
        else:
            CHKERR(DMGetLocalVector(self.dm, &vl.vec))
        CHKERR(PetscINCREF(vl.obj))
        return vl

    def restoreLocalVec(self, Vec vl, name : str | None = None) -> None:
        """Restore a local vector obtained with `getLocalVec`.

        Not collective.

        Parameters
        ----------
        vl
            The local vector.
        name
            The name used to retrieve the persistent vector, if any.

        See Also
        --------
        getLocalVec, petsc.DMRestoreLocalVector, petsc.DMRestoreNamedLocalVector

        """
        cdef const char *cname = NULL
        str2bytes(name, &cname)
        CHKERR(PetscDECREF(vl.obj))
        if cname != NULL:
            CHKERR(DMRestoreNamedLocalVector(self.dm, cname, &vl.vec))
        else:
            CHKERR(DMRestoreLocalVector(self.dm, &vl.vec))

    def globalToLocal(self, Vec vg, Vec vl, addv: InsertModeSpec | None = None) -> None:
        """Update local vectors from global vector.

        Neighborwise collective.

        Parameters
        ----------
        vg
            The global vector.
        vl
            The local vector.
        addv
            Insertion mode.

        See Also
        --------
        petsc.DMGlobalToLocalBegin, petsc.DMGlobalToLocalEnd

        """
        cdef PetscInsertMode im = insertmode(addv)
        CHKERR(DMGlobalToLocalBegin(self.dm, vg.vec, im, vl.vec))
        CHKERR(DMGlobalToLocalEnd  (self.dm, vg.vec, im, vl.vec))

    def localToGlobal(self, Vec vl, Vec vg, addv: InsertModeSpec | None = None) -> None:
        """Update global vectors from local vector.

        Neighborwise collective.

        Parameters
        ----------
        vl
            The local vector.
        vg
            The global vector.
        addv
            Insertion mode.

        See Also
        --------
        petsc.DMLocalToGlobalBegin, petsc.DMLocalToGlobalEnd

        """
        cdef PetscInsertMode im = insertmode(addv)
        CHKERR(DMLocalToGlobalBegin(self.dm, vl.vec, im, vg.vec))
        CHKERR(DMLocalToGlobalEnd(self.dm, vl.vec, im, vg.vec))

    def localToLocal(self, Vec vl, Vec vlg, addv: InsertModeSpec | None = None) -> None:
        """Map the values from a local vector to another local vector.

        Neighborwise collective.

        Parameters
        ----------
        vl
            The local vector.
        vlg
            The global vector.
        addv
            Insertion mode.

        See Also
        --------
        petsc.DMLocalToLocalBegin, petsc.DMLocalToLocalEnd

        """
        cdef PetscInsertMode im = insertmode(addv)
        CHKERR(DMLocalToLocalBegin(self.dm, vl.vec, im, vlg.vec))
        CHKERR(DMLocalToLocalEnd  (self.dm, vl.vec, im, vlg.vec))

    def getLGMap(self) -> LGMap:
        """Return local mapping to global mapping.

        Collective.

        See Also
        --------
        petsc.DMGetLocalToGlobalMapping

        """
        cdef LGMap lgm = LGMap()
        CHKERR(DMGetLocalToGlobalMapping(self.dm, &lgm.lgm))
        CHKERR(PetscINCREF(lgm.obj))
        return lgm

    #

    def getCoarseDM(self) -> DM:
        """Return the coarse `DM`.

        Collective.

        See Also
        --------
        petsc.DMGetCoarseDM

        """
        cdef DM cdm = type(self)()
        CHKERR(DMGetCoarseDM(self.dm, &cdm.dm))
        CHKERR(PetscINCREF(cdm.obj))
        return cdm

    def setCoarseDM(self, DM dm) -> None:
        """Set the coarse `DM`.

        Collective.

        See Also
        --------
        petsc.DMSetCoarseDM

        """
        CHKERR(DMSetCoarseDM(self.dm, dm.dm))
        return

    def getCoordinateDM(self) -> DM:
        """Return the coordinate `DM`.

        Collective.

        See Also
        --------
        petsc.DMGetCoordinateDM

        """
        cdef DM cdm = type(self)()
        CHKERR(DMGetCoordinateDM(self.dm, &cdm.dm))
        CHKERR(PetscINCREF(cdm.obj))
        return cdm

    def getCoordinateSection(self) -> Section:
        """Return coordinate values layout over the mesh.

        Collective.

        See Also
        --------
        petsc.DMGetCoordinateSection

        """
        cdef Section sec = Section()
        CHKERR(DMGetCoordinateSection(self.dm, &sec.sec))
        CHKERR(PetscINCREF(sec.obj))
        return sec

    def setCoordinates(self, Vec c) -> None:
        """Set a global vector that holds the coordinates.

        Collective.

        Parameters
        ----------
        c
            Coordinate Vector.

        See Also
        --------
        petsc.DMSetCoordinates

        """
        CHKERR(DMSetCoordinates(self.dm, c.vec))

    def getCoordinates(self) -> Vec:
        """Return a global vector with the coordinates associated.

        Collective.

        See Also
        --------
        petsc.DMGetCoordinates

        """
        cdef Vec c = Vec()
        CHKERR(DMGetCoordinates(self.dm, &c.vec))
        CHKERR(PetscINCREF(c.obj))
        return c

    def setCoordinatesLocal(self, Vec c) -> None:
        """Set a local vector with the ghost point holding the coordinates.

        Not collective.

        Parameters
        ----------
        c
            Coordinate Vector.

        See Also
        --------
        petsc.DMSetCoordinatesLocal

        """
        CHKERR(DMSetCoordinatesLocal(self.dm, c.vec))

    def getCoordinatesLocal(self) -> Vec:
        """Return a local vector with the coordinates associated.

        Collective the first time it is called.

        See Also
        --------
        petsc.DMGetCoordinatesLocal

        """
        cdef Vec c = Vec()
        CHKERR(DMGetCoordinatesLocal(self.dm, &c.vec))
        CHKERR(PetscINCREF(c.obj))
        return c

    def setCellCoordinateDM(self, DM dm) -> None:
        """Set the cell coordinate `DM`.

        Collective.

        Parameters
        ----------
        dm
            The cell coordinate `DM`.

        See Also
        --------
        petsc.DMSetCellCoordinateDM

        """
        CHKERR(DMSetCellCoordinateDM(self.dm, dm.dm))

    def getCellCoordinateDM(self) -> DM:
        """Return the cell coordinate `DM`.

        Collective.

        See Also
        --------
        petsc.DMGetCellCoordinateDM

        """
        cdef DM cdm = type(self)()
        CHKERR(DMGetCellCoordinateDM(self.dm, &cdm.dm))
        CHKERR(PetscINCREF(cdm.obj))
        return cdm

    def setCellCoordinateSection(self, dim: int, Section sec) -> None:
        """Set the cell coordinate layout over the `DM`.

        Collective.

        Parameters
        ----------
        dim
            The embedding dimension, or `DETERMINE`.
        sec
            The cell coordinate `Section`.

        See Also
        --------
        petsc.DMSetCellCoordinateSection

        """
        cdef PetscInt cdim = asInt(dim)
        CHKERR(DMSetCellCoordinateSection(self.dm, cdim, sec.sec))

    def getCellCoordinateSection(self) -> Section:
        """Return the cell coordinate layout over the `DM`.

        Collective.

        See Also
        --------
        petsc.DMGetCellCoordinateSection

        """
        cdef Section sec = Section()
        CHKERR(DMGetCellCoordinateSection(self.dm, &sec.sec))
        CHKERR(PetscINCREF(sec.obj))
        return sec

    def setCellCoordinates(self, Vec c) -> None:
        """Set a global vector with the cellwise coordinates.

        Collective.

        Parameters
        ----------
        c
            The global cell coordinate vector.

        See Also
        --------
        petsc.DMSetCellCoordinates

        """
        CHKERR(DMSetCellCoordinates(self.dm, c.vec))

    def getCellCoordinates(self) -> Vec:
        """Return a global vector with the cellwise coordinates.

        Collective.

        See Also
        --------
        petsc.DMGetCellCoordinates

        """
        cdef Vec c = Vec()
        CHKERR(DMGetCellCoordinates(self.dm, &c.vec))
        CHKERR(PetscINCREF(c.obj))
        return c

    def setCellCoordinatesLocal(self, Vec c) -> None:
        """Set a local vector with the cellwise coordinates.

        Not collective.

        Parameters
        ----------
        c
            The local cell coordinate vector.

        See Also
        --------
        petsc.DMSetCellCoordinatesLocal

        """
        CHKERR(DMSetCellCoordinatesLocal(self.dm, c.vec))

    def getCellCoordinatesLocal(self) -> Vec:
        """Return a local vector with the cellwise coordinates.

        Collective.

        See Also
        --------
        petsc.DMGetCellCoordinatesLocal

        """
        cdef Vec c = Vec()
        CHKERR(DMGetCellCoordinatesLocal(self.dm, &c.vec))
        CHKERR(PetscINCREF(c.obj))
        return c

    def setCoordinateDisc(self, FE disc, localized: bool, project: bool) -> Self:
        """Project coordinates to a different space.

        Collective.

        Parameters
        ----------
        disc
            The new coordinates discretization.
        localized
            Set a localized (DG) coordinate space
        project
            Project coordinates to new discretization

        See Also
        --------
        petsc.DMSetCoordinateDisc

        """
        cdef PetscBool clocalized = localized
        cdef PetscBool pr = project
        CHKERR(DMSetCoordinateDisc(self.dm, disc.fe, clocalized, pr))
        return self

    def getCoordinatesLocalized(self) -> bool:
        """Check if the coordinates have been localized for cells.

        Not collective.

        See Also
        --------
        petsc.DMGetCoordinatesLocalized

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(DMGetCoordinatesLocalized(self.dm, &flag))
        return toBool(flag)

    def getBoundingBox(self) -> tuple[tuple[float, float], ...]:
        """Return the dimension of embedding space for coordinates values.

        Not collective.

        See Also
        --------
        petsc.DMGetBoundingBox

        """
        cdef PetscInt dim=0
        CHKERR(DMGetCoordinateDim(self.dm, &dim))
        cdef PetscReal gmin[3], gmax[3]
        CHKERR(DMGetBoundingBox(self.dm, gmin, gmax))
        return tuple([(toReal(gmin[i]), toReal(gmax[i]))
                      for i from 0 <= i < dim])

    def getLocalBoundingBox(self) -> tuple[tuple[float, float], ...]:
        """Return the bounding box for the piece of the `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetLocalBoundingBox

        """
        cdef PetscInt dim=0
        CHKERR(DMGetCoordinateDim(self.dm, &dim))
        cdef PetscReal lmin[3], lmax[3]
        CHKERR(DMGetLocalBoundingBox(self.dm, lmin, lmax))
        return tuple([(toReal(lmin[i]), toReal(lmax[i]))
                      for i from 0 <= i < dim])

    def localizeCoordinates(self) -> None:
        """Create local coordinates for cells having periodic faces.

        Collective.

        Notes
        -----
        Used if the mesh is periodic.

        See Also
        --------
        petsc.DMLocalizeCoordinates

        """
        CHKERR(DMLocalizeCoordinates(self.dm))
    #

    def getPeriodicity(self) -> tuple[ArrayReal, ArrayReal, ArrayReal]:
        """Set the description of mesh periodicity.

        Not collective.

        Parameters
        ----------
        maxCell
            The longest allowable cell dimension in each direction.
        Lstart
            The start of each coordinate direction, usually [0, 0, 0]
        L
            The periodic length of each coordinate direction, or -1 for non-periodic

        See Also
        --------
        petsc.DMSetPeriodicity

        """
        cdef PetscInt dim = 0
        CHKERR(DMGetDimension(self.dm, &dim))
        cdef PetscReal *MAXCELL = NULL
        cdef ndarray mcell = oarray_r(empty_r(dim), NULL, &MAXCELL)
        cdef PetscReal *LSTART = NULL
        cdef ndarray lstart = oarray_r(empty_r(dim), NULL, &LSTART)
        cdef PetscReal *LPY = NULL
        cdef ndarray lpy = oarray_r(empty_r(dim), NULL, &LPY)
        cdef const PetscReal *maxCell = NULL
        cdef const PetscReal *Lstart = NULL
        cdef const PetscReal *L = NULL
        CHKERR(DMGetPeriodicity(self.dm, &maxCell, &Lstart, &L))
        CHKERR(PetscMemcpy(MAXCELL, maxCell, <size_t>dim*sizeof(PetscReal)))
        CHKERR(PetscMemcpy(LSTART, Lstart, <size_t>dim*sizeof(PetscReal)))
        CHKERR(PetscMemcpy(LPY, L, <size_t>dim*sizeof(PetscReal)))
        return (mcell, lstart, lpy)

    def setPeriodicity(self, maxCell: Sequence[float], Lstart: Sequence[float], L: Sequence[float]) -> None:
        """Set the description of mesh periodicity.

        Logically collective.

        Parameters
        ----------
        maxCell
            The longest allowable cell dimension in each direction.
        Lstart
            The start of each coordinate direction, usually [0, 0, 0]
        L
            The periodic length of each coordinate direction, or -1 for non-periodic

        See Also
        --------
        petsc.DMSetPeriodicity

        """
        cdef PetscReal *MAXCELL = NULL
        cdef ndarray unuseda = oarray_r(maxCell, NULL, &MAXCELL)
        cdef PetscReal *LSTART = NULL
        cdef ndarray unusedb = oarray_r(Lstart, NULL, &LSTART)
        cdef PetscReal *LPY = NULL
        cdef ndarray unusedc = oarray_r(L, NULL, &LPY)
        CHKERR(DMSetPeriodicity(self.dm, MAXCELL, LSTART, LPY))

    def setMatType(self, mat_type: Mat.Type | str) -> None:
        """Set matrix type to be used by `DM.createMat`.

        Logically collective.

        Parameters
        ----------
        mat_type
            The matrix type.

        Notes
        -----
        The option ``-dm_mat_type`` is used to set the matrix type.

        See Also
        --------
        petsc_options, petsc.DMSetMatType

        """
        cdef PetscMatType mtype = NULL
        mat_type = str2bytes(mat_type, &mtype)
        CHKERR(DMSetMatType(self.dm, mtype))

    def createMat(self) -> Mat:
        """Return an empty matrix.

        Collective.

        See Also
        --------
        petsc.DMCreateMatrix

        """
        cdef Mat mat = Mat()
        CHKERR(DMCreateMatrix(self.dm, &mat.mat))
        return mat

    def createMassMatrix(self, DM dmf) -> Mat:
        """Return the mass matrix between this `DM` and the given `DM`.

        Collective.

        Parameters
        ----------
        dmf
            The second `DM`.

        See Also
        --------
        petsc.DMCreateMassMatrix

        """
        cdef Mat mat = Mat()
        CHKERR(DMCreateMassMatrix(self.dm, dmf.dm, &mat.mat))
        return mat

    def createInterpolation(self, DM dm) -> tuple[Mat, Vec]:
        """Return the interpolation matrix to a finer `DM`.

        Collective.

        Parameters
        ----------
        dm
            The second, finer `DM`.

        See Also
        --------
        petsc.DMCreateInterpolation

        """
        cdef Mat A = Mat()
        cdef Vec scale = Vec()
        CHKERR(DMCreateInterpolation(self.dm, dm.dm,
                                     &A.mat, &scale.vec))
        return (A, scale)

    def createInjection(self, DM dm) -> Mat:
        """Return the injection matrix into a finer `DM`.

        Collective.

        Parameters
        ----------
        dm
            The second, finer `DM` object.

        See Also
        --------
        petsc.DMCreateInjection

        """
        cdef Mat inject = Mat()
        CHKERR(DMCreateInjection(self.dm, dm.dm, &inject.mat))
        return inject

    def createRestriction(self, DM dm) -> Mat:
        """Return the restriction matrix between this `DM` and the given `DM`.

        Collective.

        Parameters
        ----------
        dm
            The second, finer `DM` object.

        See Also
        --------
        petsc.DMCreateRestriction

        """
        cdef Mat mat = Mat()
        CHKERR(DMCreateRestriction(self.dm, dm.dm, &mat.mat))
        return mat

    def convert(self, dm_type: DM.Type | str) -> DM:
        """Return a `DM` converted to another `DM`.

        Collective.

        Parameters
        ----------
        dm_type
            The new `DM.Type`, use ``“same”`` for the same type.

        See Also
        --------
        DM.Type, petsc.DMConvert

        """
        cdef PetscDMType cval = NULL
        dm_type = str2bytes(dm_type, &cval)
        cdef PetscDM newdm = NULL
        CHKERR(DMConvert(self.dm, cval, &newdm))
        cdef DM dm = <DM>subtype_DM(newdm)()
        dm.dm = newdm
        return dm

    def refine(self, comm: Comm | None = None) -> DM:
        """Return a refined `DM` object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.DMRefine

        """
        cdef MPI_Comm dmcomm = MPI_COMM_NULL
        CHKERR(PetscObjectGetComm(<PetscObject>self.dm, &dmcomm))
        dmcomm = def_Comm(comm, dmcomm)
        cdef PetscDM newdm = NULL
        CHKERR(DMRefine(self.dm, dmcomm, &newdm))
        cdef DM dm = subtype_DM(newdm)()
        dm.dm = newdm
        return dm

    def coarsen(self, comm: Comm | None = None) -> DM:
        """Return a coarsened `DM` object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.DMCoarsen

        """
        cdef MPI_Comm dmcomm = MPI_COMM_NULL
        CHKERR(PetscObjectGetComm(<PetscObject>self.dm, &dmcomm))
        dmcomm = def_Comm(comm, dmcomm)
        cdef PetscDM newdm = NULL
        CHKERR(DMCoarsen(self.dm, dmcomm, &newdm))
        cdef DM dm = subtype_DM(newdm)()
        dm.dm = newdm
        return dm

    def refineHierarchy(self, nlevels: int) -> list:
        """Refine this `DM` and return the refined `DM` hierarchy.

        Collective.

        Parameters
        ----------
        nlevels
            The number of levels of refinement.

        See Also
        --------
        petsc.DMRefineHierarchy

        """
        cdef PetscInt i, n = asInt(nlevels)
        cdef PetscDM *newdmf = NULL
        cdef object unused = oarray_p(empty_p(<PetscInt>n), NULL, <void**>&newdmf)
        CHKERR(DMRefineHierarchy(self.dm, n, newdmf))
        cdef DM dmf = None
        cdef list hierarchy = []
        for i from 0 <= i < n:
            dmf = subtype_DM(newdmf[i])()
            dmf.dm = newdmf[i]
            hierarchy.append(dmf)
        return hierarchy

    def coarsenHierarchy(self, nlevels: int) -> list:
        """Coarsen this `DM` and return the coarsened `DM` hierarchy.

        Collective.

        Parameters
        ----------
        nlevels
            The number of levels of coarsening.

        See Also
        --------
        petsc.DMCoarsenHierarchy

        """
        cdef PetscInt i, n = asInt(nlevels)
        cdef PetscDM *newdmc = NULL
        cdef object unused = oarray_p(empty_p(<PetscInt>n), NULL, <void**>&newdmc)
        CHKERR(DMCoarsenHierarchy(self.dm, n, newdmc))
        cdef DM dmc = None
        cdef list hierarchy = []
        for i from 0 <= i < n:
            dmc = subtype_DM(newdmc[i])()
            dmc.dm = newdmc[i]
            hierarchy.append(dmc)
        return hierarchy

    def getRefineLevel(self) -> int:
        """Return the refinement level.

        Not collective.

        See Also
        --------
        petsc.DMGetRefineLevel

        """
        cdef PetscInt n = 0
        CHKERR(DMGetRefineLevel(self.dm, &n))
        return toInt(n)

    def setRefineLevel(self, level: int) -> None:
        """Set the number of refinements.

        Not collective.

        Parameters
        ----------
        nlevels
            The number of refinement.

        See Also
        --------
        petsc.DMSetRefineLevel

        """
        cdef PetscInt clevel = asInt(level)
        CHKERR(DMSetRefineLevel(self.dm, clevel))

    def getCoarsenLevel(self) -> int:
        """Return the number of coarsenings.

        Not collective.

        See Also
        --------
        petsc.DMGetCoarsenLevel

        """
        cdef PetscInt n = 0
        CHKERR(DMGetCoarsenLevel(self.dm, &n))
        return toInt(n)

    #

    def adaptLabel(self, label: str) -> DM:
        """Adapt a `DM` based on a `DMLabel`.

        Collective.

        Parameters
        ----------
        label
            The name of the `DMLabel`.

        See Also
        --------
        petsc.DMAdaptLabel

        """
        cdef const char *cval = NULL
        cdef PetscDMLabel clbl = NULL
        label = str2bytes(label, &cval)
        CHKERR(DMGetLabel(self.dm, cval, &clbl))
        cdef DM newdm = DMPlex()
        CHKERR(DMAdaptLabel(self.dm, clbl, &newdm.dm))
        return newdm

    def adaptMetric(
        self,
        Vec metric,
        bdLabel: str | None = None,
        rgLabel: str | None = None) -> DM:
        """Return a mesh adapted to the specified metric field.

        Collective.

        Parameters
        ----------
        metric
            The metric to which the mesh is adapted, defined vertex-wise.
        bdLabel
            Label for boundary tags.
        rgLabel
            Label for cell tag.

        See Also
        --------
        petsc.DMAdaptMetric

        """
        cdef const char *cval = NULL
        cdef PetscDMLabel cbdlbl = NULL
        cdef PetscDMLabel crglbl = NULL
        bdLabel = str2bytes(bdLabel, &cval)
        if cval == NULL: cval = b"" # XXX Should be fixed upstream
        CHKERR(DMGetLabel(self.dm, cval, &cbdlbl))
        rgLabel = str2bytes(rgLabel, &cval)
        if cval == NULL: cval = b"" # XXX Should be fixed upstream
        CHKERR(DMGetLabel(self.dm, cval, &crglbl))
        cdef DM newdm = DMPlex()
        CHKERR(DMAdaptMetric(self.dm, metric.vec, cbdlbl, crglbl, &newdm.dm))
        return newdm

    def getLabel(self, name: str) -> DMLabel:
        """Return the label of a given name.

        Not collective.

        See Also
        --------
        petsc.DMGetLabel

        """
        cdef const char *cname = NULL
        cdef DMLabel dmlabel = DMLabel()
        name = str2bytes(name, &cname)
        CHKERR(DMGetLabel(self.dm, cname, &dmlabel.dmlabel))
        CHKERR(PetscINCREF(dmlabel.obj))
        return dmlabel

    #

    def setLocalSection(self, Section sec) -> None:
        """Set the `Section` encoding the local data layout for the `DM`.

        Collective.

        See Also
        --------
        petsc.DMSetLocalSection

        """
        CHKERR(DMSetLocalSection(self.dm, sec.sec))

    def getLocalSection(self) -> Section:
        """Return the `Section` encoding the local data layout for the `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetGlobalSection

        """
        cdef Section sec = Section()
        CHKERR(DMGetLocalSection(self.dm, &sec.sec))
        CHKERR(PetscINCREF(sec.obj))
        return sec

    def setGlobalSection(self, Section sec) -> None:
        """Set the `Section` encoding the global data layout for the `DM`.

        Collective.

        See Also
        --------
        petsc.DMSetGlobalSection

        """
        CHKERR(DMSetGlobalSection(self.dm, sec.sec))

    def getGlobalSection(self) -> Section:
        """Return the `Section` encoding the global data layout for the `DM`.

        Collective the first time it is called.

        See Also
        --------
        petsc.DMGetGlobalSection

        """
        cdef Section sec = Section()
        CHKERR(DMGetGlobalSection(self.dm, &sec.sec))
        CHKERR(PetscINCREF(sec.obj))
        return sec

    setSection = setLocalSection
    getSection = getLocalSection
    setDefaultSection = setLocalSection
    getDefaultSection = getLocalSection
    setDefaultLocalSection = setLocalSection
    getDefaultLocalSection = getLocalSection
    setDefaultGlobalSection = setGlobalSection
    getDefaultGlobalSection = getGlobalSection

    def createSectionSF(self, Section localsec, Section globalsec) -> None:
        """Create the `SF` encoding the parallel DOF overlap for the `DM`.

        Collective.

        Parameters
        ----------
        localsec
            Describe the local data layout.
        globalsec
            Describe the global data layout.

        Notes
        -----
        Encoding based on the `Section` describing the data layout.

        See Also
        --------
        DM.getSectionSF, petsc.DMCreateSectionSF

        """
        CHKERR(DMCreateSectionSF(self.dm, localsec.sec, globalsec.sec))

    def getSectionSF(self) -> SF:
        """Return the `Section` encoding the parallel DOF overlap.

        Collective the first time it is called.

        See Also
        --------
        petsc.DMGetSectionSF

        """
        cdef SF sf = SF()
        CHKERR(DMGetSectionSF(self.dm, &sf.sf))
        CHKERR(PetscINCREF(sf.obj))
        return sf

    def setSectionSF(self, SF sf) -> None:
        """Set the `Section` encoding the parallel DOF overlap for the `DM`.

        Logically collective.

        See Also
        --------
        petsc.DMSetSectionSF

        """
        CHKERR(DMSetSectionSF(self.dm, sf.sf))

    createDefaultSF = createSectionSF
    getDefaultSF = getSectionSF
    setDefaultSF = setSectionSF

    def getPointSF(self) -> SF:
        """Return the `SF` encoding the parallel DOF overlap for the `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetPointSF

        """
        cdef SF sf = SF()
        CHKERR(DMGetPointSF(self.dm, &sf.sf))
        CHKERR(PetscINCREF(sf.obj))
        return sf

    def setPointSF(self, SF sf) -> None:
        """Set the `SF` encoding the parallel DOF overlap for the `DM`.

        Logically collective.

        See Also
        --------
        petsc.DMSetPointSF

        """
        CHKERR(DMSetPointSF(self.dm, sf.sf))

    def getNumLabels(self) -> int:
        """Return the number of labels defined by on the `DM`.

        Not collective.

        See Also
        --------
        petsc.DMGetNumLabels

        """
        cdef PetscInt nLabels = 0
        CHKERR(DMGetNumLabels(self.dm, &nLabels))
        return toInt(nLabels)

    def getLabelName(self, index: int) -> str:
        """Return the name of nth label.

        Not collective.

        Parameters
        ----------
        index
            The label number.

        See Also
        --------
        petsc.DMGetLabelName

        """
        cdef PetscInt cindex = asInt(index)
        cdef const char *cname = NULL
        CHKERR(DMGetLabelName(self.dm, cindex, &cname))
        return bytes2str(cname)

    def hasLabel(self, name: str) -> bool:
        """Determine whether the `DM` has a label.

        Not collective.

        Parameters
        ----------
        name
            The label name.

        See Also
        --------
        petsc.DMHasLabel

        """
        cdef PetscBool flag = PETSC_FALSE
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMHasLabel(self.dm, cname, &flag))
        return toBool(flag)

    def createLabel(self, name: str) -> None:
        """Create a label of the given name if it does not already exist.

        Not collective.

        Parameters
        ----------
        name
            The label name.

        See Also
        --------
        petsc.DMCreateLabel

        """
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMCreateLabel(self.dm, cname))

    def removeLabel(self, name: str) -> None:
        """Remove and destroy the label by name.

        Not collective.

        Parameters
        ----------
        name
            The label name.

        See Also
        --------
        petsc.DMRemoveLabel

        """
        cdef const char *cname = NULL
        cdef PetscDMLabel clbl = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMRemoveLabel(self.dm, cname, &clbl))
        # TODO: Once DMLabel is wrapped, this should return the label, like the C function.
        CHKERR(DMLabelDestroy(&clbl))

    def getLabelValue(self, name: str, point: int) -> int:
        """Return the value in `DMLabel` for the given point.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        point
            The mesh point

        See Also
        --------
        petsc.DMGetLabelValue

        """
        cdef PetscInt cpoint = asInt(point), value = 0
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMGetLabelValue(self.dm, cname, cpoint, &value))
        return toInt(value)

    def setLabelValue(self, name: str, point: int, value: int) -> None:
        """Set a point to a `DMLabel` with a given value.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        point
            The mesh point.
        value
            The label value for the point.

        See Also
        --------
        petsc.DMSetLabelValue

        """
        cdef PetscInt cpoint = asInt(point), cvalue = asInt(value)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMSetLabelValue(self.dm, cname, cpoint, cvalue))

    def clearLabelValue(self, name: str, point: int, value: int) -> None:
        """Remove a point from a `DMLabel` with given value.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        point
            The mesh point.
        value
            The label value for the point.

        See Also
        --------
        petsc.DMClearLabelValue

        """
        cdef PetscInt cpoint = asInt(point), cvalue = asInt(value)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMClearLabelValue(self.dm, cname, cpoint, cvalue))

    def getLabelSize(self, name: str) -> int:
        """Return the number of values that the `DMLabel` takes.

        Not collective.

        Parameters
        ----------
        name
            The label name.

        See Also
        --------
        petsc.DMLabelGetNumValues, petsc.DMGetLabelSize

        """
        cdef PetscInt size = 0
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMGetLabelSize(self.dm, cname, &size))
        return toInt(size)

    def getLabelIdIS(self, name: str) -> IS:
        """Return an `IS` of all values that the `DMLabel` takes.

        Not collective.

        Parameters
        ----------
        name
            The label name.

        See Also
        --------
        petsc.DMLabelGetValueIS, petsc.DMGetLabelIdIS

        """
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef IS lis = IS()
        CHKERR(DMGetLabelIdIS(self.dm, cname, &lis.iset))
        return lis

    def getStratumSize(self, name: str, value: int) -> int:
        """Return the number of points in a label stratum.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        value
            The stratum value.

        See Also
        --------
        petsc.DMGetStratumSize

        """
        cdef PetscInt size = 0
        cdef PetscInt cvalue = asInt(value)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMGetStratumSize(self.dm, cname, cvalue, &size))
        return toInt(size)

    def getStratumIS(self, name: str, value: int) -> IS:
        """Return the points in a label stratum.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        value
            The stratum value.

        See Also
        --------
        petsc.DMGetStratumIS

        """
        cdef PetscInt cvalue = asInt(value)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef IS sis = IS()
        CHKERR(DMGetStratumIS(self.dm, cname, cvalue, &sis.iset))
        return sis

    def clearLabelStratum(self, name: str, value: int) -> None:
        """Remove all points from a stratum.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        value
            The stratum value.

        See Also
        --------
        petsc.DMClearLabelStratum

        """
        cdef PetscInt cvalue = asInt(value)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        CHKERR(DMClearLabelStratum(self.dm, cname, cvalue))

    def setLabelOutput(self, name: str, output: bool) -> None:
        """Set if a given label should be saved to a view.

        Not collective.

        Parameters
        ----------
        name
            The label name.
        output
            If `True`, the label is saved to the viewer.

        See Also
        --------
        petsc.DMSetLabelOutput

        """
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscBool coutput = output
        CHKERR(DMSetLabelOutput(self.dm, cname, coutput))

    def getLabelOutput(self, name: str) -> bool:
        """Return the output flag for a given label.

        Not collective.

        Parameters
        ----------
        name
            The label name.

        See Also
        --------
        petsc.DMGetLabelOutput

        """
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscBool coutput = PETSC_FALSE
        CHKERR(DMGetLabelOutput(self.dm, cname, &coutput))
        return coutput

    # backward compatibility
    createGlobalVector = createGlobalVec
    createLocalVector = createLocalVec
    getMatrix = createMatrix = createMat

    def setKSPComputeOperators(
        self, operators,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None) -> None:
        """Matrix associated with the linear system.

        Collective.

        Parameters
        ----------
        operator
            Callback function to compute the operators.
        args
            Positional arguments for the callback.
        kargs
            Keyword arguments for the callback.

        See Also
        --------
        petsc.DMKSPSetComputeOperators

        """
        if args  is None: args  = ()
        if kargs is None: kargs = {}
        context = (operators, args, kargs)
        self.set_attr('__operators__', context)
        CHKERR(DMKSPSetComputeOperators(self.dm, KSP_ComputeOps, <void*>context))

    def createFieldDecomposition(self) -> tuple[list, list, list]:
        """Return field splitting information.

        Collective.

        See Also
        --------
        petsc.DMCreateFieldDecomposition

        """
        cdef PetscInt clen = 0
        cdef PetscIS *cis = NULL
        cdef PetscDM *cdm = NULL
        cdef char** cnamelist = NULL

        CHKERR(DMCreateFieldDecomposition(self.dm, &clen, &cnamelist, &cis, &cdm))

        cdef list isets = [ref_IS(cis[i]) for i from 0 <= i < clen]
        cdef list dms   = []
        cdef list names = []
        cdef DM dm = None

        for i from 0 <= i < clen:
            if cdm != NULL:
                dm = subtype_DM(cdm[i])()
                dm.dm = cdm[i]
                CHKERR(PetscINCREF(dm.obj))
                dms.append(dm)
                CHKERR(DMDestroy(&cdm[i]))
            else:
                dms.append(None)

            if cnamelist != NULL:
                name = bytes2str(cnamelist[i])
                names.append(name)
                CHKERR(PetscFree(cnamelist[i]))
            else:
                names.append(None)

            CHKERR(ISDestroy(&cis[i]))

        CHKERR(PetscFree(cis))
        CHKERR(PetscFree(cdm))
        CHKERR(PetscFree(cnamelist))

        return (names, isets, dms)

    def setSNESFunction(
        self, function: SNESFunction,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None) -> None:

        """Set `SNES` residual evaluation function.

        Not collective.

        Parameters
        ----------
        function
            The callback.
        args
            Positional arguments for the callback.
        kargs
            Keyword arguments for the callback.

        See Also
        --------
        SNES.setFunction, petsc.DMSNESSetFunction

        """
        if function is not None:
            if args  is None: args  = ()
            if kargs is None: kargs = {}
            context = (function, args, kargs)
            self.set_attr('__function__', context)
            CHKERR(DMSNESSetFunction(self.dm, SNES_Function, <void*>context))
        else:
            CHKERR(DMSNESSetFunction(self.dm, NULL, NULL))

    def setSNESJacobian(
        self, jacobian: SNESJacobianFunction,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None) -> None:
        """Set the `SNES` Jacobian evaluation function.

        Not collective.

        Parameters
        ----------
        jacobian
            The Jacobian callback.
        args
            Positional arguments for the callback.
        kargs
            Keyword arguments for the callback.

        See Also
        --------
        SNES.setJacobian, petsc.DMSNESSetJacobian

        """
        if jacobian is not None:
            if args  is None: args  = ()
            if kargs is None: kargs = {}
            context = (jacobian, args, kargs)
            self.set_attr('__jacobian__', context)
            CHKERR(DMSNESSetJacobian(self.dm, SNES_Jacobian, <void*>context))
        else:
            CHKERR(DMSNESSetJacobian(self.dm, NULL, NULL))

    def addCoarsenHook(
        self,
        coarsenhook: DMCoarsenHookFunction,
        restricthook: DMRestrictHookFunction,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None) -> None:
        """Add a callback to be executed when restricting to a coarser grid.

        Logically collective.

        Parameters
        ----------
        coarsenhook
            The coarsen hook function.
        restricthook
            The restrict hook function.
        args
            Positional arguments for the hooks.
        kargs
            Keyword arguments for the hooks.

        See Also
        --------
        petsc.DMCoarsenHookAdd

        """
        if args  is None: args  = ()
        if kargs is None: kargs = {}

        if coarsenhook is not None:
            coarsencontext = (coarsenhook, args, kargs)

            coarsenhooks = self.get_attr('__coarsenhooks__')
            if coarsenhooks is None:
                coarsenhooks = [coarsencontext]
                CHKERR(DMCoarsenHookAdd(self.dm, DM_PyCoarsenHook, NULL, <void*>NULL))
            else:
                coarsenhooks.append(coarsencontext)
            self.set_attr('__coarsenhooks__', coarsenhooks)

        if restricthook is not None:
            restrictcontext = (restricthook, args, kargs)

            restricthooks = self.get_attr('__restricthooks__')
            if restricthooks is None:
                restricthooks = [restrictcontext]
                CHKERR(DMCoarsenHookAdd(self.dm, NULL, DM_PyRestrictHook, <void*>NULL))
            else:
                restricthooks.append(restrictcontext)
            self.set_attr('__restricthooks__', restricthooks)

    # --- application context ---

    property appctx:
        """Application context."""
        def __get__(self) -> object:
            return self.getAppCtx()

        def __set__(self, value):
            self.setAppCtx(value)

    # --- discretization space ---

    property ds:
        """Discrete space."""
        def __get__(self) -> DS:
            return self.getDS()

        def __set__(self, value):
            self.setDS(value)

# --------------------------------------------------------------------

del DMType
del DMBoundaryType
del DMPolytopeType
del DMReorderDefaultFlag

# --------------------------------------------------------------------



import unittest
from petsc4py import PETSc

# --------------------------------------------------------------------


class TestVersion(unittest.TestCase):
    def testGetVersion(self):
        version = PETSc.Sys.getVersion()
        self.assertTrue(version > (0, 0, 0))
        v, date = PETSc.Sys.getVersion(date=True)
        self.assertTrue(version == v)
        self.assertTrue(isinstance(date, str))
        v, author = PETSc.Sys.getVersion(author=True)
        self.assertTrue(version == v)
        self.assertTrue(isinstance(author, (list, tuple)))

    def testGetVersionInfo(self):
        version = PETSc.Sys.getVersion()
        info = PETSc.Sys.getVersionInfo()
        self.assertEqual(
            version,
            (
                info['major'],
                info['minor'],
                info['subminor'],
            ),
        )
        self.assertTrue(isinstance(info['release'], bool))
        v, date = PETSc.Sys.getVersion(date=True)
        self.assertEqual(date, info['date'])

    def testGetSetDefaultComm(self):
        c = PETSc.Sys.getDefaultComm()
        self.assertEqual(c, PETSc.COMM_WORLD)
        PETSc.Sys.setDefaultComm(PETSc.COMM_SELF)
        c = PETSc.Sys.getDefaultComm()
        self.assertEqual(c, PETSc.COMM_SELF)
        PETSc.Sys.setDefaultComm(PETSc.COMM_WORLD)
        c = PETSc.Sys.getDefaultComm()
        self.assertEqual(c, PETSc.COMM_WORLD)
        f = lambda: PETSc.Sys.setDefaultComm(PETSc.COMM_NULL)
        self.assertRaises(ValueError, f)


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



# --------------------------------------------------------------------

cdef extern from * nogil:

    PetscErrorCode PetscSectionCreate(MPI_Comm, PetscSection*)
    PetscErrorCode PetscSectionClone(PetscSection, PetscSection*)
    PetscErrorCode PetscSectionSetUp(PetscSection)
    PetscErrorCode PetscSectionSetUpBC(PetscSection)
    PetscErrorCode PetscSectionView(PetscSection, PetscViewer)
    PetscErrorCode PetscSectionReset(PetscSection)
    PetscErrorCode PetscSectionDestroy(PetscSection*)

    PetscErrorCode PetscSectionGetNumFields(PetscSection, PetscInt*)
    PetscErrorCode PetscSectionSetNumFields(PetscSection, PetscInt)
    PetscErrorCode PetscSectionGetFieldName(PetscSection, PetscInt, const char*[])
    PetscErrorCode PetscSectionSetFieldName(PetscSection, PetscInt, const char[])
    PetscErrorCode PetscSectionGetFieldComponents(PetscSection, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetFieldComponents(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetChart(PetscSection, PetscInt*, PetscInt*)
    PetscErrorCode PetscSectionSetChart(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetPermutation(PetscSection, PetscIS*)
    PetscErrorCode PetscSectionSetPermutation(PetscSection, PetscIS)
    PetscErrorCode PetscSectionGetDof(PetscSection, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetDof(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionAddDof(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetFieldDof(PetscSection, PetscInt, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetFieldDof(PetscSection, PetscInt, PetscInt, PetscInt)
    PetscErrorCode PetscSectionAddFieldDof(PetscSection, PetscInt, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetConstraintDof(PetscSection, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetConstraintDof(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionAddConstraintDof(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetFieldConstraintDof(PetscSection, PetscInt, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetFieldConstraintDof(PetscSection, PetscInt, PetscInt, PetscInt)
    PetscErrorCode PetscSectionAddFieldConstraintDof(PetscSection, PetscInt, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetConstraintIndices(PetscSection, PetscInt, const PetscInt**)
    PetscErrorCode PetscSectionSetConstraintIndices(PetscSection, PetscInt, const PetscInt*)
    PetscErrorCode PetscSectionGetFieldConstraintIndices(PetscSection, PetscInt, PetscInt, const PetscInt**)
    PetscErrorCode PetscSectionSetFieldConstraintIndices(PetscSection, PetscInt, PetscInt, const PetscInt*)
    PetscErrorCode PetscSectionGetMaxDof(PetscSection, PetscInt*)
    PetscErrorCode PetscSectionGetStorageSize(PetscSection, PetscInt*)
    PetscErrorCode PetscSectionGetConstrainedStorageSize(PetscSection, PetscInt*)
    PetscErrorCode PetscSectionGetOffset(PetscSection, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetOffset(PetscSection, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetFieldOffset(PetscSection, PetscInt, PetscInt, PetscInt*)
    PetscErrorCode PetscSectionSetFieldOffset(PetscSection, PetscInt, PetscInt, PetscInt)
    PetscErrorCode PetscSectionGetOffsetRange(PetscSection, PetscInt*, PetscInt*)
    PetscErrorCode PetscSectionCreateGlobalSection(PetscSection, PetscSF, PetscBool, PetscBool, PetscBool, PetscSection*)
    PetscErrorCode PetscSectionCreateSubsection(PetscSection, PetscInt, PetscInt[], PetscSection*)
    PetscErrorCode PetscSectionCreateSubmeshSection(PetscSection, IS, PetscSection*)




cdef extern from * nogil:

    ctypedef const char* PetscSpaceType
    PetscSpaceType PETSCSPACEPOLYNOMIAL
    PetscSpaceType PETSCSPACEPTRIMMED
    PetscSpaceType PETSCSPACETENSOR
    PetscSpaceType PETSCSPACESUM
    PetscSpaceType PETSCSPACEPOINT
    PetscSpaceType PETSCSPACESUBSPACE
    PetscSpaceType PETSCSPACEWXY

    PetscErrorCode PetscSpaceCreate(MPI_Comm, PetscSpace*)
    PetscErrorCode PetscSpaceSetUp(PetscSpace)
    PetscErrorCode PetscSpaceSetFromOptions(PetscSpace)
    PetscErrorCode PetscSpaceDestroy(PetscSpace*)
    PetscErrorCode PetscSpaceView(PetscSpace, PetscViewer)
    PetscErrorCode PetscSpaceSetType(PetscSpace, PetscSpaceType)
    PetscErrorCode PetscSpaceGetType(PetscSpace, PetscSpaceType*)
    PetscErrorCode PetscSpaceGetDimension(PetscSpace, PetscInt*)
    PetscErrorCode PetscSpaceGetDegree(PetscSpace, PetscInt*, PetscInt*)
    PetscErrorCode PetscSpaceGetNumVariables(PetscSpace, PetscInt*)
    PetscErrorCode PetscSpaceGetNumComponents(PetscSpace, PetscInt*)
    PetscErrorCode PetscSpaceSetDegree(PetscSpace, PetscInt, PetscInt)
    PetscErrorCode PetscSpaceSetNumComponents(PetscSpace, PetscInt)
    PetscErrorCode PetscSpaceSetNumVariables(PetscSpace, PetscInt)

    PetscErrorCode PetscSpaceSumGetConcatenate(PetscSpace, PetscBool*)
    PetscErrorCode PetscSpaceSumSetConcatenate(PetscSpace, PetscBool)
    PetscErrorCode PetscSpaceSumGetNumSubspaces(PetscSpace, PetscInt*)
    PetscErrorCode PetscSpaceSumGetSubspace(PetscSpace, PetscInt, PetscSpace*)
    PetscErrorCode PetscSpaceSumSetNumSubspaces(PetscSpace, PetscInt)
    PetscErrorCode PetscSpaceSumSetSubspace(PetscSpace, PetscInt, PetscSpace)
    PetscErrorCode PetscSpaceTensorGetNumSubspaces(PetscSpace, PetscInt*)
    PetscErrorCode PetscSpaceTensorGetSubspace(PetscSpace, PetscInt, PetscSpace*)
    PetscErrorCode PetscSpaceTensorSetNumSubspaces(PetscSpace, PetscInt)
    PetscErrorCode PetscSpaceTensorSetSubspace(PetscSpace, PetscInt, PetscSpace)

    PetscErrorCode PetscSpacePolynomialSetTensor(PetscSpace, PetscBool)
    PetscErrorCode PetscSpacePolynomialGetTensor(PetscSpace, PetscBool*)
    PetscErrorCode PetscSpacePointSetPoints(PetscSpace, PetscQuadrature)
    PetscErrorCode PetscSpacePointGetPoints(PetscSpace, PetscQuadrature*)

    PetscErrorCode PetscSpacePTrimmedSetFormDegree(PetscSpace, PetscInt)
    PetscErrorCode PetscSpacePTrimmedGetFormDegree(PetscSpace, PetscInt*)

# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef const char* PetscDualSpaceType
    PetscDualSpaceType PETSCDUALSPACELAGRANGE
    PetscDualSpaceType PETSCDUALSPACESIMPLE
    PetscDualSpaceType PETSCDUALSPACEREFINED
    PetscDualSpaceType PETSCDUALSPACEBDM

    PetscErrorCode PetscDualSpaceCreate(MPI_Comm, PetscDualSpace*)
    PetscErrorCode PetscDualSpaceDestroy(PetscDualSpace*)
    PetscErrorCode PetscDualSpaceDuplicate(PetscDualSpace, PetscDualSpace*)
    PetscErrorCode PetscDualSpaceView(PetscDualSpace, PetscViewer)

    PetscErrorCode PetscDualSpaceGetDM(PetscDualSpace, PetscDM*)
    PetscErrorCode PetscDualSpaceSetDM(PetscDualSpace, PetscDM)
    PetscErrorCode PetscDualSpaceGetDimension(PetscDualSpace, PetscInt*)
    PetscErrorCode PetscDualSpaceGetNumComponents(PetscDualSpace, PetscInt*)
    PetscErrorCode PetscDualSpaceSetNumComponents(PetscDualSpace, PetscInt)
    PetscErrorCode PetscDualSpaceGetOrder(PetscDualSpace, PetscInt*)
    PetscErrorCode PetscDualSpaceSetOrder(PetscDualSpace, PetscInt)
    PetscErrorCode PetscDualSpaceGetNumDof(PetscDualSpace, const PetscInt**)
    PetscErrorCode PetscDualSpaceSetUp(PetscDualSpace)

    PetscErrorCode PetscDualSpaceGetFunctional(PetscDualSpace, PetscInt, PetscQuadrature*)
    PetscErrorCode PetscDualSpaceGetInteriorDimension(PetscDualSpace, PetscInt*)
    PetscErrorCode PetscDualSpaceLagrangeGetContinuity(PetscDualSpace, PetscBool*)
    PetscErrorCode PetscDualSpaceLagrangeGetTensor(PetscDualSpace, PetscBool*)
    PetscErrorCode PetscDualSpaceLagrangeGetTrimmed(PetscDualSpace, PetscBool*)
    PetscErrorCode PetscDualSpaceLagrangeSetContinuity(PetscDualSpace, PetscBool)
    PetscErrorCode PetscDualSpaceLagrangeSetTensor(PetscDualSpace, PetscBool)
    PetscErrorCode PetscDualSpaceLagrangeSetTrimmed(PetscDualSpace, PetscBool)
    PetscErrorCode PetscDualSpaceSimpleSetDimension(PetscDualSpace, PetscInt)
    PetscErrorCode PetscDualSpaceSimpleSetFunctional(PetscDualSpace, PetscInt, PetscQuadrature)
    PetscErrorCode PetscDualSpaceGetType(PetscDualSpace, PetscDualSpaceType*)
    PetscErrorCode PetscDualSpaceSetType(PetscDualSpace, PetscDualSpaceType)
